# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for the signing database models."""

import hashlib
import io
import os
import random
import re
import subprocess
from pathlib import Path
from typing import Any, BinaryIO
from unittest import mock

import gpg
from django.test import TestCase as DjangoTestCase
from gpg.results import GenkeyResult, ImportResult, ImportStatus
from nacl.public import PrivateKey

from debusine.signing.db.models import AuditLog, Key, sign
from debusine.signing.gnupg import GnuPGError, gpg_sign
from debusine.signing.models import (
    ProtectedKeyNaCl,
    ProtectedKeyPKCS11Static,
    SigningMode,
)
from debusine.signing.tests import make_fake_fingerprint
from debusine.test import TestCase


class KeyManagerTests(TestCase, DjangoTestCase):
    """Unit tests for `KeyManager`."""

    def create_nacl_key(
        self, *, storage_private_key: PrivateKey, purpose: Key.Purpose
    ) -> tuple[Key, bytes]:
        """
        Make a random NaCl-encrypted Key.

        :return: A tuple of the Key itself and its private key.
        """
        private_key = random.randbytes(64)
        public_key = random.randbytes(64)
        fingerprint = make_fake_fingerprint()
        key = Key.objects.create(
            purpose=purpose,
            fingerprint=fingerprint,
            private_key=ProtectedKeyNaCl.encrypt(
                storage_private_key.public_key, private_key
            ).model_dump(),
            public_key=public_key,
        )
        return key, private_key

    def test_generate_uefi(self) -> None:
        """Generate a UEFI key."""
        # Generating keys is slow, so we just test that the correct
        # subprocesses are executed.
        private_key = random.randbytes(64)
        public_key = random.randbytes(64)
        fingerprint = make_fake_fingerprint()
        storage_private_key = PrivateKey.generate()
        log_file = io.BytesIO()

        def run(
            args: list[str | os.PathLike[str]], **kwargs: Any
        ) -> subprocess.CompletedProcess[Any]:
            match args[1]:
                case "req":
                    Path(args[args.index("-keyout") + 1]).write_bytes(
                        private_key
                    )
                    Path(args[args.index("-out") + 1]).write_bytes(public_key)
                    return subprocess.CompletedProcess(args=args, returncode=0)
                case "x509":
                    fp_colons = ":".join(re.findall("..", fingerprint))
                    return subprocess.CompletedProcess(
                        args=args,
                        returncode=0,
                        stdout=f"sha256 Fingerprint={fp_colons}\n".encode(),
                    )
                case _:  # pragma: no cover
                    return subprocess.CompletedProcess(args=args, returncode=1)

        with (
            mock.patch("subprocess.run", side_effect=run) as mock_run,
            self.settings(DEBUSINE_SIGNING_PRIVATE_KEYS=[storage_private_key]),
        ):
            key = Key.objects.generate(
                Key.Purpose.UEFI, "debian/bookworm test key", 123, log_file
            )

        self.assertEqual(mock_run.call_count, 2)
        mock_run.assert_has_calls(
            [
                mock.call(
                    [
                        "openssl",
                        "req",
                        "-new",
                        "-newkey",
                        "rsa:2048",
                        "-x509",
                        "-subj",
                        r"/CN=debian\/bookworm test key/",
                        "-days",
                        "5475",
                        "-noenc",
                        "-sha256",
                        "-keyout",
                        mock.ANY,
                        "-out",
                        mock.ANY,
                    ],
                    check=True,
                    stdout=log_file,
                    stderr=log_file,
                ),
                mock.call(
                    [
                        "openssl",
                        "x509",
                        "-inform",
                        "PEM",
                        "-noout",
                        "-fingerprint",
                        "-sha256",
                    ],
                    check=True,
                    input=mock.ANY,
                    stdout=subprocess.PIPE,
                    stderr=log_file,
                ),
            ]
        )
        self.assertEqual(key.purpose, Key.Purpose.UEFI)
        self.assertEqual(key.fingerprint, fingerprint)
        assert isinstance(key.stored_private_key.root, ProtectedKeyNaCl)
        self.assertEqual(
            key.stored_private_key.root.decrypt([storage_private_key]),
            private_key,
        )
        self.assertEqual(key.public_key, public_key)
        audit_log = AuditLog.objects.latest("created_at")
        self.assertEqual(audit_log.purpose, Key.Purpose.UEFI)
        self.assertEqual(audit_log.fingerprint, fingerprint)
        self.assertEqual(audit_log.event, AuditLog.Event.GENERATE)
        self.assertEqual(
            audit_log.data, {"description": "debian/bookworm test key"}
        )
        self.assertEqual(audit_log.created_by_work_request_id, 123)

    def test_generate_openpgp(self) -> None:
        """Generate an OpenPGP key."""
        # Generating keys is slow, so we mock most of this and just test the
        # sequence of operations.  Integration tests will make sure that it
        # actually works.
        secret_key = random.randbytes(64)
        public_key = random.randbytes(64)
        fingerprint = make_fake_fingerprint(length=40)
        create_result = GenkeyResult(None)
        create_result.primary = True
        create_result.sub = False
        create_result.fpr = fingerprint
        import_status = ImportStatus(None)
        import_status.fpr = fingerprint
        import_result = ImportResult(None)
        import_result.imported = 1
        import_result.imports = [import_status]
        storage_private_key = PrivateKey.generate()
        log_file = io.BytesIO()

        for purpose in (
            Key.Purpose.OPENPGP_UPLOAD,
            Key.Purpose.OPENPGP_REPOSITORY,
        ):
            with self.subTest(purpose=purpose.value):
                with (
                    mock.patch(
                        "gpg.Context.create_key", return_value=create_result
                    ) as mock_create_key,
                    mock.patch(
                        "gpg.Context.key_export_secret", return_value=secret_key
                    ) as mock_key_export_secret,
                    mock.patch(
                        "gpg.Context.key_export_minimal",
                        return_value=public_key,
                    ) as mock_key_export_minimal,
                    mock.patch(
                        "gpg.Context.key_import", return_value=import_result
                    ) as mock_key_import,
                    self.settings(
                        DEBUSINE_SIGNING_PRIVATE_KEYS=[storage_private_key]
                    ),
                ):
                    key = Key.objects.generate(
                        purpose, "test key", 123, log_file
                    )

                mock_create_key.assert_called_once_with(
                    "test key", algorithm="ed25519", expires=False, sign=True
                )
                mock_key_export_secret.assert_called_once_with(fingerprint)
                mock_key_export_minimal.assert_called_once_with(fingerprint)
                mock_key_import.assert_called_once_with(public_key)
                self.assertEqual(key.purpose, purpose)
                self.assertEqual(key.fingerprint, fingerprint)
                assert isinstance(key.stored_private_key.root, ProtectedKeyNaCl)
                self.assertEqual(
                    key.stored_private_key.root.decrypt([storage_private_key]),
                    secret_key,
                )
                self.assertEqual(key.public_key, public_key)
                audit_log = AuditLog.objects.latest("created_at")
                self.assertEqual(audit_log.purpose, purpose)
                self.assertEqual(audit_log.fingerprint, fingerprint)
                self.assertEqual(audit_log.event, AuditLog.Event.GENERATE)
                self.assertEqual(audit_log.data, {"description": "test key"})
                self.assertEqual(audit_log.created_by_work_request_id, 123)

    def test_generate_openpgp_create_key_no_primary(self) -> None:
        """Key generation error: failed to generate an OpenPGP secret key."""
        create_result = GenkeyResult(None)
        create_result.primary = False
        log_file = io.BytesIO()

        with (
            mock.patch("gpg.Context.create_key", return_value=create_result),
            self.assertRaisesRegex(
                GnuPGError, "Failed to generate a secret key"
            ),
        ):
            Key.objects.generate(
                Key.Purpose.OPENPGP_UPLOAD, "test key", 123, log_file
            )

    def test_generate_openpgp_create_key_unexpected_subkey(self) -> None:
        """Key generation error: unexpectedly got OpenPGP encryption subkey."""
        create_result = GenkeyResult(None)
        create_result.primary = True
        create_result.sub = True
        log_file = io.BytesIO()

        with (
            mock.patch("gpg.Context.create_key", return_value=create_result),
            self.assertRaisesRegex(
                GnuPGError, "Unexpectedly got an encryption subkey"
            ),
        ):
            Key.objects.generate(
                Key.Purpose.OPENPGP_UPLOAD, "test key", 123, log_file
            )

    def test_generate_openpgp_key_export_secret_failed(self) -> None:
        """Key generation error: failed to export OpenPGP secret key."""
        fingerprint = make_fake_fingerprint(length=40)
        create_result = GenkeyResult(None)
        create_result.primary = True
        create_result.sub = False
        create_result.fpr = fingerprint
        log_file = io.BytesIO()

        with (
            mock.patch("gpg.Context.create_key", return_value=create_result),
            mock.patch("gpg.Context.key_export_secret", return_value=None),
            self.assertRaisesRegex(GnuPGError, "Failed to export secret key"),
        ):
            Key.objects.generate(
                Key.Purpose.OPENPGP_UPLOAD, "test key", 123, log_file
            )

    def test_generate_openpgp_key_export_public_failed(self) -> None:
        """Key generation error: failed to export OpenPGP public key."""
        secret_key = random.randbytes(64)
        fingerprint = make_fake_fingerprint(length=40)
        create_result = GenkeyResult(None)
        create_result.primary = True
        create_result.sub = False
        create_result.fpr = fingerprint
        log_file = io.BytesIO()

        with (
            mock.patch("gpg.Context.create_key", return_value=create_result),
            mock.patch(
                "gpg.Context.key_export_secret", return_value=secret_key
            ),
            mock.patch("gpg.Context.key_export_minimal", return_value=None),
            self.assertRaisesRegex(GnuPGError, "Failed to export public key"),
        ):
            Key.objects.generate(
                Key.Purpose.OPENPGP_UPLOAD, "test key", 123, log_file
            )

    def test_generate_openpgp_key_import_failed(self) -> None:
        """Key generation error: failed to import OpenPGP public key."""
        secret_key = random.randbytes(64)
        public_key = random.randbytes(64)
        fingerprint = make_fake_fingerprint(length=40)
        create_result = GenkeyResult(None)
        create_result.primary = True
        create_result.sub = False
        create_result.fpr = fingerprint
        import_result = ImportResult(None)
        import_result.not_imported = 1
        log_file = io.BytesIO()

        with (
            mock.patch("gpg.Context.create_key", return_value=create_result),
            mock.patch(
                "gpg.Context.key_export_secret", return_value=secret_key
            ),
            mock.patch(
                "gpg.Context.key_export_minimal", return_value=public_key
            ),
            mock.patch("gpg.Context.key_import", return_value=import_result),
            self.assertRaisesRegex(
                GnuPGError, "Failed to get fingerprint of new key"
            ),
        ):
            Key.objects.generate(
                Key.Purpose.OPENPGP_UPLOAD, "test key", 123, log_file
            )

    def test_sign_uefi_attached(self) -> None:
        """Sign data using a UEFI key, returning an attached signature."""
        # Signing data is slow, so we just test that the correct
        # subprocesses are executed.
        storage_private_key = PrivateKey.generate()
        key, private_key = self.create_nacl_key(
            storage_private_key=storage_private_key, purpose=Key.Purpose.UEFI
        )
        temp_path = self.create_temporary_directory()
        raw_data = b"data to sign"
        (data_path := temp_path / "data").write_bytes(raw_data)
        signature_path = temp_path / "signed"
        signature_bytes = random.randbytes(64)
        log_file = io.BytesIO()
        key_bytes: bytes | None = None
        certificate_bytes: bytes | None = None

        def run(
            args: list[str | os.PathLike[str]], **kwargs: Any
        ) -> subprocess.CompletedProcess[Any]:
            nonlocal key_bytes, certificate_bytes
            key_bytes = Path(args[args.index("--key") + 1]).read_bytes()
            certificate_bytes = Path(
                args[args.index("--cert") + 1]
            ).read_bytes()
            Path(args[args.index("--output") + 1]).write_bytes(signature_bytes)
            return subprocess.CompletedProcess(args=args, returncode=0)

        with (
            mock.patch("subprocess.run", side_effect=run) as mock_run,
            self.settings(DEBUSINE_SIGNING_PRIVATE_KEYS=[storage_private_key]),
        ):
            sign(
                [key],
                data_path,
                signature_path,
                SigningMode.ATTACHED,
                123,
                username="testuser",
                user_id=1234,
                resource={"package": "linux"},
                log_file=log_file,
            )

        mock_run.assert_called_once_with(
            [
                "sbsign",
                "--key",
                mock.ANY,
                "--cert",
                mock.ANY,
                "--output",
                signature_path,
                data_path,
            ],
            check=True,
            stdout=log_file,
            stderr=log_file,
        )
        self.assertEqual(key_bytes, private_key)
        self.assertEqual(certificate_bytes, key.public_key)
        self.assertEqual(signature_path.read_bytes(), signature_bytes)

        audit_log = AuditLog.objects.latest("created_at")
        self.assertEqual(audit_log.purpose, Key.Purpose.UEFI)
        self.assertEqual(audit_log.fingerprint, key.fingerprint)
        self.assertEqual(audit_log.event, AuditLog.Event.SIGN)
        self.assertEqual(
            audit_log.data,
            {
                "data_sha256": hashlib.sha256(raw_data).hexdigest(),
                "username": "testuser",
                "user_id": 1234,
                "resource": {
                    "package": "linux",
                },
            },
        )
        self.assertEqual(audit_log.created_by_work_request_id, 123)

    def test_sign_uefi_detached(self) -> None:
        """Sign data using a UEFI key, returning a detached signature."""
        # Signing data is slow, so we just test that the correct
        # subprocesses are executed.
        storage_private_key = PrivateKey.generate()
        key, private_key = self.create_nacl_key(
            storage_private_key=storage_private_key, purpose=Key.Purpose.UEFI
        )
        temp_path = self.create_temporary_directory()
        (data_path := temp_path / "data").write_bytes(b"data to sign")
        signature_path = temp_path / "signed"
        signature_bytes = random.randbytes(64)
        log_file = io.BytesIO()
        key_bytes: bytes | None = None
        certificate_bytes: bytes | None = None

        def run(
            args: list[str | os.PathLike[str]], **kwargs: Any
        ) -> subprocess.CompletedProcess[Any]:
            nonlocal key_bytes, certificate_bytes
            key_bytes = Path(args[args.index("--key") + 1]).read_bytes()
            certificate_bytes = Path(
                args[args.index("--cert") + 1]
            ).read_bytes()
            Path(args[args.index("--output") + 1]).write_bytes(signature_bytes)
            return subprocess.CompletedProcess(args=args, returncode=0)

        with (
            mock.patch("subprocess.run", side_effect=run) as mock_run,
            self.settings(DEBUSINE_SIGNING_PRIVATE_KEYS=[storage_private_key]),
        ):
            sign(
                [key],
                data_path,
                signature_path,
                SigningMode.DETACHED,
                123,
                log_file=log_file,
            )

        mock_run.assert_called_once_with(
            [
                "sbsign",
                "--key",
                mock.ANY,
                "--cert",
                mock.ANY,
                "--detached",
                "--output",
                signature_path,
                data_path,
            ],
            check=True,
            stdout=log_file,
            stderr=log_file,
        )
        self.assertEqual(key_bytes, private_key)
        self.assertEqual(certificate_bytes, key.public_key)
        self.assertEqual(signature_path.read_bytes(), signature_bytes)

    def test_sign_openpgp_clear(self) -> None:
        """Sign data using an OpenPGP key, returning a clear signature."""
        # Signing data is slow, so we mock most of this and just test the
        # sequence of operations.  Integration tests will make sure that it
        # actually works.
        storage_private_key = PrivateKey.generate()
        for purpose in (
            Key.Purpose.OPENPGP_UPLOAD,
            Key.Purpose.OPENPGP_REPOSITORY,
        ):
            with self.subTest(purpose=purpose.value):
                key, secret_key = self.create_nacl_key(
                    storage_private_key=storage_private_key, purpose=purpose
                )
                import_status = ImportStatus(None)
                import_status.fpr = key.fingerprint
                import_result = ImportResult(None)
                import_result.secret_imported = 1
                import_result.imports = [import_status]
                keylist = [object()]
                temp_path = self.create_temporary_directory()
                data_bytes = b"data to sign"
                (data_path := temp_path / "data").write_bytes(data_bytes)
                signature_path = temp_path / "data.asc"
                signature_bytes = random.randbytes(64)

                def gpg_sign(
                    ctx: gpg.Context,
                    data: gpg.Data,
                    sink: BinaryIO | None = None,
                    mode: int = gpg.constants.SIG_MODE_NORMAL,
                ) -> None:
                    self.assertTrue(ctx.armor)
                    self.assertEqual(data.read(), data_bytes)
                    self.assertEqual(mode, gpg.constants.SIG_MODE_CLEAR)
                    assert sink is not None
                    sink.write(signature_bytes)

                with (
                    self.settings(
                        DEBUSINE_SIGNING_PRIVATE_KEYS=[storage_private_key]
                    ),
                    mock.patch(
                        "gpg.Context.key_import", return_value=import_result
                    ) as mock_key_import,
                    mock.patch(
                        "gpg.Context.keylist", return_value=iter(keylist)
                    ) as mock_keylist,
                    mock.patch(
                        "gpg.Context.signers", new_callable=mock.PropertyMock
                    ) as mock_signers,
                    mock.patch(
                        "gpg.Context.sign", autospec=True, side_effect=gpg_sign
                    ) as mock_sign,
                ):
                    sign(
                        [key], data_path, signature_path, SigningMode.CLEAR, 123
                    )

                mock_key_import.assert_called_once_with(secret_key)
                mock_keylist.assert_called_once_with(
                    pattern=key.fingerprint, secret=True
                )
                mock_signers.assert_called_with(keylist)
                mock_sign.assert_called_once()
                self.assertEqual(signature_path.read_bytes(), signature_bytes)

    def test_sign_openpgp_detached(self) -> None:
        """Sign data using an OpenPGP key, returning a detached signature."""
        # Signing data is slow, so we mock most of this and just test the
        # sequence of operations.  Integration tests will make sure that it
        # actually works.
        storage_private_key = PrivateKey.generate()
        for purpose in (
            Key.Purpose.OPENPGP_UPLOAD,
            Key.Purpose.OPENPGP_REPOSITORY,
        ):
            with self.subTest(purpose=purpose.value):
                key, secret_key = self.create_nacl_key(
                    storage_private_key=storage_private_key, purpose=purpose
                )
                import_status = ImportStatus(None)
                import_status.fpr = key.fingerprint
                import_result = ImportResult(None)
                import_result.secret_imported = 1
                import_result.imports = [import_status]
                keylist = [object()]
                temp_path = self.create_temporary_directory()
                data_bytes = b"data to sign"
                (data_path := temp_path / "data").write_bytes(data_bytes)
                signature_path = temp_path / "data.asc"
                signature_bytes = random.randbytes(64)

                def gpg_sign(
                    ctx: gpg.Context,
                    data: gpg.Data,
                    sink: BinaryIO | None = None,
                    mode: int = gpg.constants.SIG_MODE_NORMAL,
                ) -> None:
                    self.assertTrue(ctx.armor)
                    self.assertEqual(data.read(), data_bytes)
                    self.assertEqual(mode, gpg.constants.SIG_MODE_DETACH)
                    assert sink is not None
                    sink.write(signature_bytes)

                with (
                    self.settings(
                        DEBUSINE_SIGNING_PRIVATE_KEYS=[storage_private_key]
                    ),
                    mock.patch(
                        "gpg.Context.key_import", return_value=import_result
                    ) as mock_key_import,
                    mock.patch(
                        "gpg.Context.keylist", return_value=iter(keylist)
                    ) as mock_keylist,
                    mock.patch(
                        "gpg.Context.signers", new_callable=mock.PropertyMock
                    ) as mock_signers,
                    mock.patch(
                        "gpg.Context.sign", autospec=True, side_effect=gpg_sign
                    ) as mock_sign,
                ):
                    sign(
                        [key],
                        data_path,
                        signature_path,
                        SigningMode.DETACHED,
                        123,
                    )

                mock_key_import.assert_called_once_with(secret_key)
                mock_keylist.assert_called_once_with(
                    pattern=key.fingerprint, secret=True
                )
                mock_signers.assert_called_with(keylist)
                mock_sign.assert_called_once()
                self.assertEqual(signature_path.read_bytes(), signature_bytes)

    def test_sign_openpgp_multiple(self) -> None:
        """Sign data using multiple OpenPGP keys."""
        # Signing data is slow, so we mock most of this and just test the
        # sequence of operations.  Integration tests will make sure that it
        # actually works.
        storage_private_key = PrivateKey.generate()
        key_pairs = [
            self.create_nacl_key(
                storage_private_key=storage_private_key,
                purpose=Key.Purpose.OPENPGP_REPOSITORY,
            )
            for _ in range(2)
        ]
        import_results: list[ImportResult] = []
        for key, _ in key_pairs:
            import_status = ImportStatus(None)
            import_status.fpr = key.fingerprint
            import_result = ImportResult(None)
            import_result.secret_imported = 1
            import_result.imports = [import_status]
            import_results.append(import_result)
        keylist = [object() for _ in key_pairs]
        temp_path = self.create_temporary_directory()
        data_bytes = b"data to sign"
        (data_path := temp_path / "data").write_bytes(data_bytes)
        signature_path = temp_path / "data.asc"
        signature_bytes = random.randbytes(64)

        def gpg_sign(
            ctx: gpg.Context,
            data: gpg.Data,
            sink: BinaryIO | None = None,
            mode: int = gpg.constants.SIG_MODE_NORMAL,
        ) -> None:
            self.assertTrue(ctx.armor)
            self.assertEqual(data.read(), data_bytes)
            self.assertEqual(mode, gpg.constants.SIG_MODE_CLEAR)
            assert sink is not None
            sink.write(signature_bytes)

        with (
            self.settings(DEBUSINE_SIGNING_PRIVATE_KEYS=[storage_private_key]),
            mock.patch(
                "gpg.Context.key_import", side_effect=import_results
            ) as mock_key_import,
            mock.patch(
                "gpg.Context.keylist",
                side_effect=[iter([signer]) for signer in keylist],
            ) as mock_keylist,
            mock.patch(
                "gpg.Context.signers", new_callable=mock.PropertyMock
            ) as mock_signers,
            mock.patch(
                "gpg.Context.sign", autospec=True, side_effect=gpg_sign
            ) as mock_sign,
        ):
            sign(
                [key for key, _ in key_pairs],
                data_path,
                signature_path,
                SigningMode.CLEAR,
                123,
            )

        self.assertEqual(mock_key_import.call_count, len(key_pairs))
        mock_key_import.assert_has_calls(
            [mock.call(secret_key) for _, secret_key in key_pairs]
        )
        self.assertEqual(mock_keylist.call_count, len(key_pairs))
        mock_keylist.assert_has_calls(
            [
                mock.call(pattern=key.fingerprint, secret=True)
                for key, _ in key_pairs
            ]
        )
        mock_signers.assert_called_with(keylist)
        mock_sign.assert_called_once()
        self.assertEqual(signature_path.read_bytes(), signature_bytes)

    def test_sign_openpgp_not_detached_or_clear(self) -> None:
        """Key signing error: signing mode not detached or clear."""
        storage_private_key = PrivateKey.generate()
        key, _ = self.create_nacl_key(
            storage_private_key=storage_private_key,
            purpose=Key.Purpose.OPENPGP_UPLOAD,
        )
        temp_path = self.create_temporary_directory()
        (data_path := temp_path / "data").write_bytes(b"data to sign")
        signature_path = temp_path / "data.asc"

        with self.assertRaisesRegex(
            GnuPGError,
            "OpenPGP signing currently only supports detached or clear "
            "signatures",
        ):
            # Bypass debusine.signing.db.models.sign, since it also catches
            # this.
            gpg_sign(
                [(key.stored_private_key, key.public_key)],
                data_path,
                signature_path,
                SigningMode.ATTACHED,
            )

    def test_sign_openpgp_not_software_encrypted(self) -> None:
        """Key signing error: OpenPGP key is not software-encrypted."""
        public_key = random.randbytes(64)
        fingerprint = make_fake_fingerprint()
        key = Key.objects.create(
            purpose=Key.Purpose.OPENPGP_UPLOAD,
            fingerprint=fingerprint,
            private_key=ProtectedKeyPKCS11Static.create(
                pkcs11_uri="pkcs11:id=1"
            ).model_dump(),
            public_key=public_key,
        )
        temp_path = self.create_temporary_directory()
        (data_path := temp_path / "data").write_bytes(b"data to sign")
        signature_path = temp_path / "data.asc"

        with self.assertRaisesRegex(
            GnuPGError,
            "OpenPGP signing currently only supports software-encrypted keys",
        ):
            sign([key], data_path, signature_path, SigningMode.CLEAR, 123)

    def test_sign_openpgp_key_import_failed(self) -> None:
        """Key signing error: failed to import OpenPGP secret key."""
        # Signing data is slow, so we mock most of this and just test the
        # sequence of operations.  Integration tests will make sure that it
        # actually works.
        storage_private_key = PrivateKey.generate()
        key, _ = self.create_nacl_key(
            storage_private_key=storage_private_key,
            purpose=Key.Purpose.OPENPGP_UPLOAD,
        )
        import_status = ImportStatus(None)
        import_status.fpr = key.fingerprint
        import_result = ImportResult(None)
        import_result.not_imported = 1
        temp_path = self.create_temporary_directory()
        (data_path := temp_path / "data").write_bytes(b"data to sign")
        signature_path = temp_path / "data.asc"

        with (
            self.settings(DEBUSINE_SIGNING_PRIVATE_KEYS=[storage_private_key]),
            mock.patch("gpg.Context.key_import", return_value=import_result),
            self.assertRaisesRegex(GnuPGError, "Failed to import secret key"),
        ):
            sign([key], data_path, signature_path, SigningMode.CLEAR, 123)

    def test_sign_openpgp_debsign(self) -> None:
        """Sign an upload using an OpenPGP key and debsign."""
        # Signing data is slow, so we mock most of this and just test the
        # sequence of operations.  Integration tests will make sure that it
        # actually works.
        storage_private_key = PrivateKey.generate()
        key, secret_key = self.create_nacl_key(
            storage_private_key=storage_private_key,
            purpose=Key.Purpose.OPENPGP_UPLOAD,
        )
        import_status = ImportStatus(None)
        import_status.fpr = key.fingerprint
        import_result = ImportResult(None)
        import_result.secret_imported = 1
        import_result.imports = [import_status]
        unsigned_path = self.create_temporary_directory()
        (unsigned_path / "foo.dsc").write_text("dsc\n")
        (unsigned_changes_path := unsigned_path / "foo.changes").write_text(
            "changes\n"
        )
        signed_path = self.create_temporary_directory()
        signed_dsc_path = signed_path / "foo.dsc"
        signed_changes_path = signed_path / "foo.changes"
        signature_dsc = "signed dsc\n"
        signature_changes = "signed changes\n"
        log_file = io.BytesIO()

        def run(
            args: list[str | os.PathLike[str]], **kwargs: Any
        ) -> subprocess.CompletedProcess[Any]:
            if args[0] == "debsign":
                changes_path = Path(args[-1])
                dsc_path = changes_path.parent / f"{changes_path.stem}.dsc"
                dsc_path.write_text(signature_dsc)
                changes_path.write_text(signature_changes)
            return subprocess.CompletedProcess(args=args, returncode=0)

        with (
            self.settings(DEBUSINE_SIGNING_PRIVATE_KEYS=[storage_private_key]),
            mock.patch(
                "gpg.Context.key_import", return_value=import_result
            ) as mock_key_import,
            mock.patch("subprocess.run", side_effect=run) as mock_run,
        ):
            sign(
                [key],
                unsigned_changes_path,
                signed_changes_path,
                SigningMode.DEBSIGN,
                123,
                log_file=log_file,
            )

        mock_key_import.assert_called_once_with(secret_key)
        mock_run.assert_any_call(
            [
                "debsign",
                f"-k{key.fingerprint}",
                "--re-sign",
                signed_changes_path,
            ],
            cwd=signed_changes_path.parent,
            env={**os.environ, "GNUPGHOME": mock.ANY},
            check=True,
            stdout=log_file,
            stderr=log_file,
        )
        self.assertEqual(signed_dsc_path.read_text(), signature_dsc)
        self.assertEqual(signed_changes_path.read_text(), signature_changes)

    def test_sign_openpgp_debsign_not_software_encrypted(self) -> None:
        """Key signing error: OpenPGP key is not software-encrypted."""
        public_key = random.randbytes(64)
        fingerprint = make_fake_fingerprint()
        key = Key.objects.create(
            purpose=Key.Purpose.OPENPGP_UPLOAD,
            fingerprint=fingerprint,
            private_key=ProtectedKeyPKCS11Static.create(
                pkcs11_uri="pkcs11:id=1"
            ).model_dump(),
            public_key=public_key,
        )
        unsigned_path = self.create_temporary_directory()
        (unsigned_changes_path := unsigned_path / "foo.changes").write_text(
            "changes\n"
        )
        signed_path = self.create_temporary_directory()
        signed_changes_path = signed_path / "foo.changes"

        with self.assertRaisesRegex(
            GnuPGError,
            "OpenPGP signing currently only supports software-encrypted keys",
        ):
            sign(
                [key],
                unsigned_changes_path,
                signed_changes_path,
                SigningMode.DEBSIGN,
                123,
            )

    def test_sign_openpgp_debsign_key_import_failed(self) -> None:
        """Key signing error: failed to import OpenPGP secret key."""
        # Signing data is slow, so we mock most of this and just test the
        # sequence of operations.  Integration tests will make sure that it
        # actually works.
        storage_private_key = PrivateKey.generate()
        key, _ = self.create_nacl_key(
            storage_private_key=storage_private_key,
            purpose=Key.Purpose.OPENPGP_UPLOAD,
        )
        import_status = ImportStatus(None)
        import_status.fpr = key.fingerprint
        import_result = ImportResult(None)
        import_result.not_imported = 1
        unsigned_path = self.create_temporary_directory()
        (unsigned_changes_path := unsigned_path / "foo.changes").write_text(
            "changes\n"
        )
        signed_path = self.create_temporary_directory()
        signed_changes_path = signed_path / "foo.changes"

        with (
            self.settings(DEBUSINE_SIGNING_PRIVATE_KEYS=[storage_private_key]),
            mock.patch("gpg.Context.key_import", return_value=import_result),
            self.assertRaisesRegex(GnuPGError, "Failed to import secret key"),
        ):
            sign(
                [key],
                unsigned_changes_path,
                signed_changes_path,
                SigningMode.DEBSIGN,
                123,
            )

    def test_sign_openpgp_debsign_signature_directory_not_empty(self) -> None:
        """Key signing error: directory containing signature_path not empty."""
        # Signing data is slow, so we mock most of this and just test the
        # sequence of operations.  Integration tests will make sure that it
        # actually works.
        storage_private_key = PrivateKey.generate()
        key, _ = self.create_nacl_key(
            storage_private_key=storage_private_key,
            purpose=Key.Purpose.OPENPGP_UPLOAD,
        )
        import_status = ImportStatus(None)
        import_status.fpr = key.fingerprint
        import_result = ImportResult(None)
        import_result.secret_imported = 1
        import_result.imports = [import_status]
        unsigned_path = self.create_temporary_directory()
        (unsigned_changes_path := unsigned_path / "foo.changes").write_text(
            "changes\n"
        )
        signed_path = self.create_temporary_directory()
        signed_changes_path = signed_path / "foo.changes"
        (signed_path / "foo.dsc").touch()

        with (
            self.settings(DEBUSINE_SIGNING_PRIVATE_KEYS=[storage_private_key]),
            mock.patch("gpg.Context.key_import", return_value=import_result),
            self.assertRaisesRegex(
                GnuPGError, "Directory containing signature_path is not empty"
            ),
        ):
            sign(
                [key],
                unsigned_changes_path,
                signed_changes_path,
                SigningMode.DEBSIGN,
                123,
            )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for scheduler tags backend functions."""

import ast
from unittest import TestCase

import pyparsing

from debusine.utils.tag_utils import (
    CandidateTags,
    DerivationRuleParser,
    DerivationRules,
    ProvenanceRestrictions,
)


class ProvenanceRestrictionsTests(TestCase):
    """Tests for :py:class:`ProvenanceRestrictions`."""

    def test_name(self) -> None:
        r = ProvenanceRestrictions("foo")
        self.assertEqual(r.name, "foo")

    def test_add_exact(self) -> None:
        r = ProvenanceRestrictions("foo")
        r.add_exact("foo", ["bar"])
        r.add_exact("wibble", {"wobble"})
        self.assertEqual(
            r.exact,
            {
                "foo": {"bar"},
                "wibble": {"wobble"},
            },
        )
        self.assertEqual(r.prefixes, [])

    def test_add_exact_twice(self) -> None:
        r = ProvenanceRestrictions("prov")
        r.add_exact("foo", ["bar"])
        with self.assertRaisesRegex(
            ValueError,
            r"prov: exact match for tag foo specified multiple times",
        ):
            r.add_exact("foo", ["bar"])

    def test_add_prefix(self) -> None:
        r = ProvenanceRestrictions("foo")
        r.add_prefix("foo", ("bar",))
        r.add_prefix("wibble", ["wobble"])
        self.assertEqual(r.exact, {})
        self.assertEqual(
            r.prefixes,
            [
                ("foo", {"bar"}),
                ("wibble", {"wobble"}),
            ],
        )

    def test_filter_set(self) -> None:
        r = ProvenanceRestrictions("test")
        r.add_prefix("group", ["system"])
        r.add_exact("official:foo", ["system", "scope:foo"])
        r.add_exact("official:bar", ["system", "scope:bar"])
        r.add_prefix("official:bar", ["scope:bar"])
        for provenance, tags, expected in (
            ("foo", set(), True),
            ("foo", {"foo", "bar"}, True),
            ("system", {"group:foo", "group:bar", "baz"}, True),
            ("other", {"group:foo", "group:bar", "baz"}, {"baz"}),
            (
                "system",
                {"group:foo", "official:foo", "official:bar"},
                {"group:foo", "official:foo"},
            ),
            (
                "scope:foo",
                {"group:foo", "official:foo", "official:bar"},
                {"official:foo"},
            ),
            (
                "scope:bar",
                {"group:foo", "official:foo", "official:bar"},
                {"official:bar"},
            ),
        ):
            with self.subTest(provenance=provenance, tags=repr(tags)):
                if expected is True:
                    expected = tags
                self.assertEqual(r.filter_set(provenance, tags), expected)


class DerivationRuleParserTests(TestCase):
    """Tests for :py:class:`DerivationRuleParser`."""

    def test_tokens_invalid_string(self) -> None:
        parser = DerivationRuleParser()
        for value, pos in (
            ("/", 0),
            (" /", 1),
            ("tag@tag", 3),
            ("[tag]", 0),
            ("tåg", 1),
        ):
            with (
                self.subTest(value=value),
                self.assertRaisesRegex(
                    pyparsing.ParseException,
                    rf"Expected .*, found '.*'  \(at char {pos}\)",
                ),
            ):
                parser.as_expr(value)

    def test_as_expr(self) -> None:
        parser = DerivationRuleParser()
        for value, parsed in (
            ("tag", "'tag' in tags"),
            ("not tag", "not 'tag' in tags"),
            ("(tag)", "'tag' in tags"),
            # See https://github.com/pyparsing/pyparsing/issues/204
            ("(((((((((tag)))))))))", "'tag' in tags"),
            ("   (  tag  )   ", "'tag' in tags"),
            ("tag1 and tag2", "'tag1' in tags and 'tag2' in tags"),
            (
                "tag1 and tag2 and tag3",
                "'tag1' in tags and 'tag2' in tags and ('tag3' in tags)",
            ),
            ("tag1 or tag2", "'tag1' in tags or 'tag2' in tags"),
            (
                "tag1 or tag2 or tag3",
                "'tag1' in tags or 'tag2' in tags or 'tag3' in tags",
            ),
            (
                "tag1 and tag2 or tag3",
                "'tag1' in tags and 'tag2' in tags or 'tag3' in tags",
            ),
            (
                "tag and (tag1 or tag2)",
                "'tag' in tags and ('tag1' in tags or 'tag2' in tags)",
            ),
            (
                "(tag1 and not tag2) or tag3",
                "'tag1' in tags and (not 'tag2' in tags) or 'tag3' in tags",
            ),
            (
                "tag1 or tag2 and tag3",
                "'tag1' in tags or ('tag2' in tags and 'tag3' in tags)",
            ),
            ("not tag1 and tag2", "not 'tag1' in tags and 'tag2' in tags"),
            (
                "not not not (((tag1))) or tag3",
                "not not not 'tag1' in tags or 'tag3' in tags",
            ),
            (
                "tag1 and not tag2 and tag3",
                "'tag1' in tags and (not 'tag2' in tags) and ('tag3' in tags)",
            ),
            (
                "tag1 and (tag2 or tag3)",
                "'tag1' in tags and ('tag2' in tags or 'tag3' in tags)",
            ),
            ("fooandbar", "'fooandbar' in tags"),
            ("fooorbar", "'fooorbar' in tags"),
        ):
            with self.subTest(value=value):
                expr = parser.as_expr(value)
                self.assertEqual(ast.unparse(expr), parsed)

    def test_as_expr_invalid_syntax(self) -> None:
        parser = DerivationRuleParser()
        for expr, message in (
            ("foo(", r"Expected end of text, found '\('"),
            ("foo bar", "Expected end of text, found 'bar'"),
            ("foo and", "Expected end of text, found 'and'"),
            ("or bar", r"Expected 'or' (term|operations), found 'or'"),
            ("()", r"Expected 'or' (term|operations), found '\)'"),
            ("(or)", "Expected 'or' (term|operations), found 'or'"),
            ("(and)", "Expected 'or' (term|operations), found 'and'"),
            ("foo.__globals__", r"Expected end of text, found '\.'"),
            ("foo + bar", r"Expected end of text, found '\+'"),
            ("' or \"", "Expected 'or' (term|operations), found \"'\""),
        ):
            with (
                self.subTest(expr=expr),
                self.assertRaisesRegex(pyparsing.ParseException, message),
            ):
                parser.as_expr(expr)

    def test_as_function(self) -> None:
        parser = DerivationRuleParser()
        for value, tags, result in (
            ("foo", {"foo", "bar"}, True),
            ("foo", {"bar"}, False),
            ("foo and bar", {"foo", "bar"}, True),
            ("foo and bar", {"foo"}, False),
            ("foo or bar", {"foo", "bar"}, True),
            ("foo or bar", {"foo"}, True),
            ("not foo", {"bar"}, True),
            ("not foo", {"foo"}, False),
        ):
            f = parser.as_function(value, "test")
            self.assertEqual(f(tags), result)


class DerivationRulesTests(TestCase):
    """Tests for :py:class:`DerivationRules`."""

    def test_provenance(self) -> None:
        r = DerivationRules(provenance="foo")
        self.assertEqual(r.provenance, "foo")

    def test_compute(self) -> None:
        r = DerivationRules(provenance="test")
        r.add_rule("dd or dm", {"debian"})
        r.add_rule("(dd or dm) and dsa", {"embargoed"})
        r.add_rule("debian", {"official"})
        for tags, expected in (
            (set(), set()),
            ({"foo"}, set()),
            ({"dd"}, {"debian"}),
            ({"dm"}, {"debian"}),
            ({"dd", "dsa"}, {"debian", "embargoed"}),
            ({"dm", "dsa"}, {"debian", "embargoed"}),
            ({"foo", "dsa"}, set()),
            ({"foo", "debian"}, {"official"}),
        ):
            with self.subTest(tags=repr(tags)):
                self.assertEqual(r.compute(tags), expected)


class CandidateTagsTests(TestCase):
    """Tests for :py:class:`CandidateTags`."""

    def test_empty(self) -> None:
        c: CandidateTags[str] = CandidateTags()
        self.assertEqual(c.sets, {})

    def test_add(self) -> None:
        c: CandidateTags[str] = CandidateTags()
        c.add("p1", {"a", "b"})
        c.add("p2", {"a", "c"})
        c.add("p1", {"a", "c"})
        self.assertEqual(c.sets, {"p1": {"a", "b", "c"}, "p2": {"a", "c"}})

    def test_add_candidates(self) -> None:
        c: CandidateTags[str] = CandidateTags()
        c.add("p1", {"a", "b"})
        c.add("p2", {"a", "c"})

        c1: CandidateTags[str] = CandidateTags()
        c1.add("p1", {"a", "c"})
        c1.add("p2", {"a", "d"})
        c1.add("p3", {"e"})

        c.add_candidates(c1)

        self.assertEqual(
            c.sets, {"p1": {"a", "b", "c"}, "p2": {"a", "c", "d"}, "p3": {"e"}}
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Debusine extensions to Python's standard unittest.TestCase."""

import contextlib
import datetime as dt
import hashlib
import inspect
import io
import os
import pkgutil
import re
import shutil
import tempfile
import textwrap
import unittest
from collections import defaultdict
from collections.abc import Generator, Mapping
from configparser import ConfigParser
from pathlib import Path
from typing import Any, ClassVar, Optional, TYPE_CHECKING, cast, override
from unittest import mock
from unittest.util import safe_repr

import debian.deb822 as deb822
import lxml
import lxml.etree
import lxml.objectify
import responses
import yaml

from debusine.artifacts.local_artifact import deb822dict_to_dict
from debusine.artifacts.models import DebianSourcePackage
from debusine.artifacts.playground import ArtifactPlayground

if TYPE_CHECKING:
    from unittest._log import _LoggingWatcher

    from _typeshed import ReadableBuffer

    ReadableBuffer  # fake usage for vulture


class TestCase(unittest.TestCase):
    """
    Collection of methods to help write unit tests.

    This augments unittest.TestCase with assert statements and factory
    functions to support Debusine unit tests.

    This is intended to be used as the TestCase for tests that do not depend on
    Django code.
    """

    _initial_cls_precious_globals: ClassVar[dict[str, Any]]
    _initial_self_precious_globals: dict[str, Any]

    # The unittest defaults mostly just get in the way of debugging.
    maxDiff = None

    @classmethod
    def get_precious_globals(cls) -> dict[str, Any]:
        """Collect global values that tests should not change."""
        return {"cwd": os.getcwd()}

    @override
    @classmethod
    def setUpClass(cls) -> None:
        """Save the values of globals that should not be changed by tests."""
        cls._initial_cls_precious_globals = cls.get_precious_globals()
        super().setUpClass()

    @override
    @classmethod
    def tearDownClass(cls) -> None:
        """Check that saved globals haven't been changed by tests."""
        super().tearDownClass()
        if not hasattr(cls, "_initial_cls_precious_globals"):
            raise AssertionError(
                f"{inspect.getfile(cls)}:"
                f"{cls.__qualname__}.setUpClass"
                " did not call super().setUpClass"
            )
        if cls._initial_cls_precious_globals != cls.get_precious_globals():
            raise AssertionError(
                f"{inspect.getfile(cls)}:"
                f"{cls.__qualname__}:"
                f" {cls._initial_cls_precious_globals}"
                f" != {cls.get_precious_globals()}"
            )

    @override
    def setUp(self) -> None:
        """Save the values of globals that should not be changed by tests."""
        self._initial_self_precious_globals = self.get_precious_globals()
        super().setUp()

    @override
    def tearDown(self) -> None:
        """Check that saved globals haven't been changed by tests."""
        super().tearDown()
        initial = getattr(self, "_initial_self_precious_globals", None)
        self.assertIsNotNone(initial, "setUp did not call super().setUp")
        self.assertEqual(
            self._initial_self_precious_globals,
            self.get_precious_globals(),
        )

    def create_temp_config_directory(self, config: dict[str, Any]) -> Path:
        """
        Create a temp directory with a config.ini file inside.

        The method also register the automatic removal of said directory.
        """
        temp_directory = Path(tempfile.mkdtemp())
        config_file_name = temp_directory / 'config.ini'
        with config_file_name.open('w') as config_file:
            config_writer = ConfigParser()

            for section, values in config.items():
                config_writer[section] = values

            config_writer.write(config_file)

        self.addCleanup(shutil.rmtree, temp_directory)

        return temp_directory

    def assertDictContainsAll(
        self,
        dictionary: Mapping[Any, Any],
        subset: Mapping[Any, Any],
        msg: Any = None,
    ) -> None:
        """
        Implement a replacement of deprecated TestCase.assertDictContainsSubset.

        Assert that the keys and values of subset is in dictionary.

        The order of the arguments in TestCase.assertDictContainsSubset
        and this implementation differs.
        """
        self.assertIsInstance(
            dictionary, Mapping, 'First argument is not a mapping'
        )
        self.assertIsInstance(
            subset, Mapping, 'Second argument is not a mapping'
        )

        if dictionary != dict(dictionary) | dict(subset):
            msg = self._formatMessage(
                msg,
                f"{safe_repr(dictionary)} does not contain the subset"
                f" {safe_repr(subset)}",
            )

            raise self.failureException(msg)

    def assert_token_key_included_in_all_requests(
        self, expected_token: str
    ) -> None:
        """Assert that the requests in responses.calls had the Token."""
        for call in responses.calls:
            headers = call.request.headers

            if 'Token' not in headers:
                raise self.failureException(
                    'Token missing in the headers for '
                    f'the request {safe_repr(call.request.url)}'
                )

            if (actual_token := headers['Token']) != expected_token:
                raise self.failureException(
                    f"Unexpected token."
                    f" In the request: {safe_repr(call.request.url)}"
                    f" Actual: {safe_repr(actual_token)}"
                    f" Expected: {safe_repr(expected_token)}"
                )

    def create_temporary_file(
        self,
        *,
        prefix: str | None = None,
        suffix: str | None = None,
        contents: Optional["ReadableBuffer"] = None,
        directory: Path | str | None = None,
    ) -> Path:
        """
        Create a temporary file and schedules the deletion via self.addCleanup.

        :param prefix: prefix is "debusine-tests-" + prefix or "debusine-tests-"
        :param suffix: suffix for the created file
        :param contents: contents is written into the file. If it's none the
          file is left empty
        :param directory: the directory that the file is created into.
        """
        if prefix is None:
            prefix = "debusine-tests-"
        else:
            prefix = "debusine-tests-" + prefix

        suffix = suffix or ""

        with tempfile.NamedTemporaryFile(
            prefix=f"{prefix}-", suffix=suffix, delete=False, dir=directory
        ) as file:
            if contents is not None:
                file.write(contents)

        file_path = Path(file.name)

        self.addCleanup(file_path.unlink, missing_ok=True)

        return file_path

    def create_temporary_directory(
        self, *, directory: Path | str | None = None
    ) -> Path:
        """
        Create and return a temporary directory. Schedules deletion.

        :param directory: directory to create the temporary directory. If None,
          use the default for tempfile.TemporaryDirectory.
        """
        temp_dir = tempfile.TemporaryDirectory(
            prefix="debusine-tests-", dir=directory
        )

        self.addCleanup(temp_dir.cleanup)

        return Path(temp_dir.name)

    @contextlib.contextmanager
    def assertRaisesSystemExit(self, exit_code: int) -> Generator[None]:
        """Assert that raises SystemExit with the specific exit_code."""
        with self.assertRaisesRegex(
            SystemExit,
            rf'^{exit_code}$',
            msg=f'Did not raise SystemExit with exit_code=^{exit_code}$',
        ):
            yield

    @contextlib.contextmanager
    def assertLogsContains(
        self, message: str, expected_count: int = 1, **assert_logs_kwargs: Any
    ) -> "Generator[_LoggingWatcher]":
        """
        Raise failureException if message is not in the logs.

        Yields the same context manager as self.assertLogs(). This allows
        further checks in the logs.

        :param message: message to find in the logs
        :param expected_count: expected times that the message
          must be in the logs
        :param assert_logs_kwargs: arguments for self.assertLogs()
        """

        def failure_exception_if_needed(
            logs: "_LoggingWatcher", message: str, expected_count: int
        ) -> None:
            all_logs = '\n'.join(logs.output)

            actual_times = all_logs.count(message)

            if actual_times != expected_count:
                raise self.failureException(
                    f'Expected: "{message}"\n'
                    f'Actual: "{all_logs}"\n'
                    f'Expected msg found {actual_times} times,'
                    f' expected {expected_count} times'
                )

        with self.assertLogs(**assert_logs_kwargs) as logs:
            try:
                yield logs
            except BaseException as exc:
                failure_exception_if_needed(logs, message, expected_count)
                raise exc

        failure_exception_if_needed(logs, message, expected_count)

    @staticmethod
    def write_dsc_example_file(path: Path) -> dict[str, Any]:
        """Write a DSC file into file. Files in .dsc are not created."""
        metadata = {
            "source": "hello",
            "version": "2.10-2",
        }
        text = textwrap.dedent(
            f"""\
            -----BEGIN PGP SIGNED MESSAGE-----
            Hash: SHA256

            Format: 3.0 (quilt)
            Source: {metadata['source']}
            Binary: hello
            Architecture: any
            Version: {metadata['version']}
            Maintainer: Santiago Vila <sanvila@debian.org>
            Homepage: http://www.gnu.org/software/hello/
            Standards-Version: 4.3.0
            Build-Depends: debhelper-compat (= 9)
            Package-List:
             hello deb devel optional arch=any
            Checksums-Sha1:
             f7bebf6f9c62a2295e889f66e05ce9bfaed9ace3 725946 hello_2.10.orig.tar.gz
             a35d97bd364670b045cdd86d446e71b171e915cc 6132 hello_2.10-2.debian.tar.xz
            Checksums-Sha256:
             31e066137a962676e89f69d1b65382de95a7ef7d914b8cb956f41ea72e0f516b 725946 hello_2.10.orig.tar.gz
             811ad0255495279fc98dc75f4460da1722f5c1030740cb52638cb80d0fdb24f0 6132 hello_2.10-2.debian.tar.xz
            Files:
             6cd0ffea3884a4e79330338dcc2987d6 725946 hello_2.10.orig.tar.gz
             e522e61c27eb0401c86321b9d8e137ae 6132 hello_2.10-2.debian.tar.xz

            -----BEGIN PGP SIGNATURE-----
            """  # noqa: E501
        )
        path.write_text(text)

        dsc = deb822.Dsc(text.encode("utf-8"))

        return deb822dict_to_dict(dsc)

    @classmethod
    def write_dsc_file(
        cls, path: Path, files: list[Path], version: str = "2.10-5"
    ) -> dict[str, Any]:
        """Write a debian .dsc file."""
        return ArtifactPlayground.write_deb822_file(
            deb822.Dsc, path, files, version=version
        )

    def write_deb_file(
        self,
        path: Path,
        *,
        source_name: str | None = None,
        source_version: str | None = None,
        control_file_names: list[str] | None = None,
    ) -> dict[str, Any]:
        """Write a debian control file."""
        return ArtifactPlayground.write_deb_file(
            path,
            source_name=source_name,
            source_version=source_version,
            control_file_names=control_file_names,
        )

    @classmethod
    def write_changes_file(
        cls,
        path: Path,
        files: list[Path],
        binaries: list[str] | None = None,
        version: str = "2.10-5",
        binnmu: bool = False,
    ) -> dict[str, Any]:
        """Write a debian .changes file."""
        return ArtifactPlayground.write_deb822_file(
            deb822.Changes,
            path,
            files,
            binaries=binaries,
            version=version,
            binnmu=binnmu,
        )

    def mock_is_command_available(self, commands: dict[str, bool]) -> None:
        """
        Configure a fake is_command_available.

        It responds by looking up the requested command in the given
        `commands` dictionary.
        """
        patcher = mock.patch(
            "debusine.utils.is_command_available",
            side_effect=lambda cmd: commands.get(cmd, False),
        )
        patcher.start()
        self.addCleanup(patcher.stop)

    def assert_source_artifact_equal(
        self,
        source: DebianSourcePackage,
        name: str = "hello",
        version: str = "1.0-1",
    ) -> None:
        """
        Verify the contents of a source package artifact.

        File contents are assumed to have been autogenerated using playground
        defaults.
        """
        files: list[str]
        if "-" in version:
            files = [
                f"{name}_{version}.debian.tar.xz",
                f"{name}_{version.split('-')[0]}.orig.tar.gz",
            ]
        else:
            files = [
                f"{name}_{version}.tar.gz",
            ]
        self.assertEqual(source.name, name)
        self.assertEqual(source.version, version)
        self.assertEqual(source.type, "dpkg")

        def csum(name: str, text: str) -> str:
            hasher = hashlib.new(name)
            hasher.update(text.encode())
            return hasher.hexdigest()

        checksums_sha1: list[dict[str, str]] = [
            {"name": name, "sha1": csum("sha1", name), "size": str(len(name))}
            for name in files
        ]
        checksums_sha256: list[dict[str, str]] = [
            {
                "name": name,
                "sha256": csum("sha256", name),
                "size": str(len(name)),
            }
            for name in files
        ]
        checksums_md5: list[dict[str, str]] = [
            {"name": name, "md5sum": csum("md5", name), "size": str(len(name))}
            for name in files
        ]
        package_list = source.dsc_fields.pop("Package-List")
        self.assertEqual(
            package_list,
            [
                {
                    "package": name,
                    "package-type": "deb",
                    "section": "devel",
                    "priority": "optional",
                    "_other": "arch=any",
                }
            ],
        )
        self.assertEqual(
            source.dsc_fields,
            {
                'Architecture': 'any',
                'Binary': 'hello',
                'Checksums-Sha1': checksums_sha1,
                'Checksums-Sha256': checksums_sha256,
                'Files': checksums_md5,
                'Format': '3.0 (quilt)',
                'Homepage': f'http://www.gnu.org/software/{name}/',
                'Maintainer': 'Example Maintainer <example@example.org>',
                'Source': name,
                'Standards-Version': '4.3.0',
                'Version': version,
            },
        )

    # LXML does not seem to know about HTML5 structural tags
    re_lxml_false_positive_tags = re.compile(
        r"Tag (?:nav|footer|header|article|details|summary) invalid"
    )

    def _filter_parser_error_log(
        self, error_log: lxml.etree._ListErrorLog
    ) -> list[str]:
        """Filter lxml parser error log for known false positives."""
        # _LogEntry documentation:
        # https://lxml.de/apidoc/lxml.etree.html#lxml.etree._LogEntry
        Domains = lxml.etree.ErrorDomains

        errors: list[str] = []
        for error in error_log:
            match (error.domain, error.type_name):
                # Nice to have if available, but libxml2 >= 2.14.0 no longer
                # provides unknown-tag checks at all so we can't reliably
                # cover this.
                case Domains.HTML, "HTML_UNKNOWN_TAG":  # pragma: no cover
                    if self.re_lxml_false_positive_tags.match(error.message):
                        continue

            errors.append(f"{error.line}:{error.type_name}:{error.message}")
        return errors

    def dump_html(
        self,
        content: str,
        *,
        title: str | None = None,
        parser: lxml.etree.HTMLParser | None = None,
    ) -> None:
        """
        Dump HTML on standard output.

        :param content: HTML string to dump
        :param title: optional title text to write before the dump
        :param parser: optional parser instance that parsed the HTML, to use to
                       annotate code with parser errors
        """
        from rich.console import Console
        from rich.syntax import Syntax

        console = Console()
        if title is not None:
            console.print("")
            console.print(
                f"    {title}",
                markup=False,
                highlight=False,
                style="bold bright_red",
            )
            console.print("")

        errors_by_line: dict[int, list[str]] | None = None
        if parser:
            errors_by_line = defaultdict(list)
            for e in parser.error_log:
                errors_by_line[e.line].append(f"{e.type_name}:{e.message}")
        else:
            errors_by_line = {}

        syntax = Syntax(content, "html", line_numbers=True)
        text = syntax.highlight(content)
        for lineno, line in enumerate(text.split(), start=1):
            console.print(
                f"{lineno:03d}", end=" ", style="bright_black", highlight=False
            )
            console.print(line)
            if line_errors := errors_by_line.get(lineno):
                for error in line_errors:
                    console.print(
                        f"  → {error}",
                        style="red",
                        markup=False,
                        highlight=False,
                    )

    def dump_element(
        self,
        element: lxml.objectify.ObjectifiedElement,
        *,
        title: str | None = None,
    ) -> None:
        """
        Dump a LXML element on standard output.

        :param element: element to dump
        :param title: optional title text to write before the dump
        """
        self.dump_html(lxml.etree.tostring(element).decode(), title=title)

    def assertHTMLValid(
        self,
        content: str | bytes,
        dump_on_error: bool = False,
        check_widget_errors: bool = True,
    ) -> lxml.objectify.ObjectifiedElement:
        """
        Parse the response contents as HTML and checks contents.

        Assert:
        - HTML parses without errors
        - No "debusine-template-errors" in the contents

        Returns the parsed tree.
        """
        if isinstance(content, str):
            content = content.encode()
            parser = lxml.etree.HTMLParser(
                remove_blank_text=True, encoding="utf8"
            )
        else:
            parser = lxml.etree.HTMLParser(remove_blank_text=True)
        parser.set_element_class_lookup(
            lxml.objectify.ObjectifyElementClassLookup()
        )
        with io.BytesIO(content) as fd:
            tree = lxml.etree.parse(fd, parser)
        errors = self._filter_parser_error_log(parser.error_log)

        # Dump source with errors if requested
        if errors and dump_on_error:
            self.dump_html(
                content.decode(), parser=parser, title="HTML parser errors"
            )

        self.assertEqual(errors, [])

        if check_widget_errors:
            # Assert that no "debusine-template-error" is displayed. It appears
            # if a widget has an error
            self.assertEqual(
                tree.xpath('//*[@data-role="debusine-template-error"]'), []
            )

        return cast(lxml.objectify.ObjectifiedElement, tree.getroot())

    def assertHasElement(
        self,
        tree: lxml.objectify.ObjectifiedElement,
        xpath: str,
        dump_on_error: bool = False,
    ) -> lxml.objectify.ObjectifiedElement:
        """
        Ensure that the tree has strictly one matching element.

        :param tree: tree to search
        :param xpath: XPath query to match
        :returns: resulting element
        """
        el = tree.xpath(xpath)
        if not el:
            message = f"{xpath!r} not found in tree"
            if dump_on_error:
                self.dump_element(tree, title=message)
            self.fail(message)
        if len(el) > 1:
            message = f"{xpath!r} matched {len(el)} elements instead of one"
            if dump_on_error:
                self.dump_element(tree, title=message)
            self.fail(message)
        assert isinstance(el[0], lxml.objectify.ObjectifiedElement)
        return el[0]

    def assertElementHasClass(
        self, element: lxml.objectify.ObjectifiedElement, name: str
    ) -> None:
        """Ensure the element has the given class in the DOM."""
        dom_class = element.get("class")
        if dom_class is None:
            self.fail("Element does not have a class attribute")
        self.assertIn(name, dom_class.split())

    def assertElementHasNoClass(
        self, element: lxml.objectify.ObjectifiedElement, name: str
    ) -> None:
        """Ensure the element does not have the given class in the DOM."""
        if (dom_class := element.get("class")) is None:
            return
        self.assertNotIn(name, dom_class.split())

    def get_node_text_normalized(
        self, node: lxml.objectify.ObjectifiedElement
    ) -> str:
        """Get the node text contents, with spaces normalized."""
        # itertext will iterate on whitespace-only blocks, so it needs two
        # passes: one to reconstruct the text, and one to normalise whitespace
        sample = "".join(node.itertext())
        return " ".join(sample.strip().split())

    def assertTextContentEqual(
        self,
        node: lxml.objectify.ObjectifiedElement,
        text: str,
        dump_on_error: bool = False,
    ) -> None:
        """
        Ensure that node.text matches the given text.

        Both expected and actual text are normalised so that consecutive
        whitespace and newlines become a single space, to simplify dealing with
        the way HTML collapses whitespace.
        """
        sample = self.get_node_text_normalized(node)
        text = " ".join(text.strip().split())
        if dump_on_error and sample != text:
            self.dump_element(node, title="Text content mismatch")
        self.assertEqual(sample, text)

    def assertDatetimeContentEqual(
        self,
        node: lxml.objectify.ObjectifiedElement,
        d: dt.datetime,
        anywhere: bool = False,
    ) -> None:
        """
        Ensure the row contains a datetime matching the given one.

        Text surrounding the datetime value is ignored.
        """
        text = self.get_node_text_normalized(node)
        regex = r"\d+-\d+-\d+ \d+:\d+"
        if not anywhere:
            regex = f"^{regex}$"

        if not (m := re.search(regex, text)):
            self.fail(
                f"text content {text!r} does not contain"
                " a YYYY-MM-DD HH:MM datetime"
            )
        actual = dt.datetime.strptime(m.group(), "%Y-%m-%d %H:%M").replace(
            tzinfo=dt.UTC
        )
        expected = d.replace(second=0, microsecond=0)
        self.assertEqual(actual, expected)

    def assertYAMLContentEqual(
        self, node: lxml.objectify.ObjectifiedElement, data: Any
    ) -> None:
        """Ensure that the node contents match yaml-formatted data."""
        encoded = "".join(node.itertext())
        try:
            actual = yaml.safe_load(encoded)
        except yaml.parser.ParserError as e:
            self.fail(f"node text {encoded!r} does not parse as YAML: {e}")
        self.assertEqual(actual, data)

    @contextlib.contextmanager
    def spy(
        self, target: str, *, side_effect: Any | None = None
    ) -> Generator[mock.MagicMock]:
        """
        Wrap the given function with a mock object.

        Calling the function will work as usual, but the mock object can be
        used to inspect its calls.
        """
        # From mock._get_target
        try:
            target_obj_name, attribute = target.rsplit('.', 1)
        except (TypeError, ValueError, AttributeError):
            raise TypeError(
                f"Need a valid target to patch. You supplied: {target!r}"
            )
        target_obj = pkgutil.resolve_name(target_obj_name)
        wrapped = getattr(target_obj, attribute)

        with mock.patch(
            target, autospec=True, wraps=wrapped, side_effect=side_effect
        ) as mocked:
            yield mocked

    @contextlib.contextmanager
    def spy_object(
        self, obj: Any, name: str, *, side_effect: Any | None = None
    ) -> Generator[mock.MagicMock]:
        """
        Wrap an object member with a mock object.

        Calling the member will work as usual, but the mock object can be used
        to inspect its calls.
        """
        with mock.patch.object(
            obj,
            name,
            autospec=True,
            wraps=getattr(obj, name),
            side_effect=side_effect,
        ) as mocked:
            yield mocked



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Utilities used by debusine.signing."""

from pathlib import Path
from tempfile import TemporaryDirectory
from typing import AnyStr, override

from django.core.exceptions import ImproperlyConfigured
from nacl.exceptions import CryptoError
from nacl.public import PrivateKey


# TODO: The signing worker's systemd unit must use something along the lines
# of "TemporaryFileSystem=/tmp:noswap /var/tmp:noswap" so that private keys
# don't end up on disk.  The "noswap" option was added in Linux 6.4, so we'd
# either have to skip it and accept the possibility that private keys might
# be swapped to disk under memory pressure, or use a backported kernel for
# signing workers.
class SensitiveTemporaryDirectory(TemporaryDirectory[AnyStr]):
    """
    A temporary directory that may contain sensitive data.

    This class does nothing special right now, but it provides a marker for
    places where we might need to add additional safety checks.
    """

    @override
    def __repr__(self) -> str:
        """
        Adjust repr to note that this instance contains sensitive data.

        This has the effect that resource warnings about instances of this
        class not being cleaned up correctly will have something noticeable
        in their text.
        """
        return super().__repr__() + " (SENSITIVE)"


def read_private_key(private_key_path: Path) -> PrivateKey:
    """Read a NaCl private key."""
    try:
        if private_key_path.stat().st_mode & 0o077:
            raise ImproperlyConfigured(
                f"Permission too open for {private_key_path}. "
                f"Make sure that the file is not accessible by group or others."
            )
        private_key_bytes = private_key_path.read_bytes()
    except OSError as e:
        raise ImproperlyConfigured(f"Cannot read {private_key_path}: {e}")
    try:
        return PrivateKey(private_key_bytes)
    except CryptoError as e:
        raise ImproperlyConfigured(
            f"Cannot load key from {private_key_path}: {e}"
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for the database aware task field resolver."""

import re
from collections.abc import Callable
from types import SimpleNamespace
from typing import Any, ClassVar, cast, override
from unittest import mock

from django.test import override_settings
from pgpy import PGPKey, PGPKeyring
from pgpy.constants import EllipticCurveOID, PubKeyAlgorithm

from debusine.artifacts.models import ArtifactCategory, CollectionCategory
from debusine.assets.asset_utils import ExportablePGPKeyring
from debusine.assets.models import KeyPurpose
from debusine.client.models import LookupChildType, RelationType
from debusine.db.context import context
from debusine.db.models import (
    Artifact,
    ArtifactRelation,
    Collection,
    WorkRequest,
)
from debusine.db.permissioncontext import PermissionContext
from debusine.db.playground import scenarios
from debusine.db.tasks import DBWorkerProxyTask
from debusine.db.tasks.models import DBArtifactInfo, DBBaseTaskInputArtifact
from debusine.db.tasks.resolver import DBFieldResolver
from debusine.tasks import BaseTaskWithExecutor, TaskConfigError
from debusine.tasks.executors import ExecutorImageCategory
from debusine.tasks.inputs import FieldResolver, inputs
from debusine.tasks.models import (
    BackendType,
    ExtraDebusineRepository,
    ExtraExternalRepository,
    LookupMultiple,
)
from debusine.test.django import TestCase
from debusine.test.test_utils import create_system_tarball_data


class DBFieldResolverTests(TestCase):
    """Test :py:class:`DBFieldResolver`."""

    scenario = scenarios.DefaultContext()
    work_request: ClassVar[WorkRequest]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.work_request = cls.playground.create_worker_task(
            task_name="noop", workspace=cls.scenario.workspace
        )

    def resolver(self) -> DBFieldResolver:
        """Instantiate the DBFieldResolver to test."""
        task = self.work_request.get_task()
        return DBFieldResolver(task)

    @context.disable_permission_checks()
    def test_lookup_single_artifact(self) -> None:
        """Look up a single artifact."""
        collection = self.playground.create_collection(
            "debian",
            CollectionCategory.ENVIRONMENTS,
            workspace=self.scenario.workspace,
        )
        artifact, _ = self.playground.create_artifact(
            category=ArtifactCategory.SYSTEM_TARBALL,
            data=create_system_tarball_data(
                codename="bookworm", architecture="amd64"
            ),
        )
        collection.manager.add_artifact(
            artifact, user=self.playground.get_default_user()
        )
        resolver = self.resolver()

        lookup = "debian@debian:environments/match:codename=bookworm"
        result = resolver.lookup_single_artifact(lookup, label="test")
        self.assertEqual(
            result,
            DBArtifactInfo.from_artifact(artifact, label="test", lookup=lookup),
        )
        with self.assertNumQueries(0):
            self.assertEqual(result.artifact, artifact)

        lookup = "debian/match:codename=bookworm"
        result = resolver.lookup_single_artifact(
            lookup,
            default_category=CollectionCategory.ENVIRONMENTS,
            label="test",
        )
        self.assertEqual(
            result,
            DBArtifactInfo.from_artifact(artifact, lookup=lookup, label="test"),
        )
        with self.assertNumQueries(0):
            self.assertEqual(result.artifact, artifact)
        with self.assertRaisesRegex(
            LookupError,
            "'debian' does not specify a category and the context does not "
            "supply a default",
        ):
            resolver.lookup_single_artifact(
                "debian/match:codename=bookworm", label="test"
            )

    def test_lookup_multiple_artifacts(self) -> None:
        collection = self.playground.create_collection(
            "debian",
            CollectionCategory.ENVIRONMENTS,
            workspace=self.scenario.workspace,
        )
        artifacts: list[Artifact] = []
        for codename in "bookworm", "trixie":
            artifact, _ = self.playground.create_artifact(
                category=ArtifactCategory.SYSTEM_TARBALL,
                data=create_system_tarball_data(
                    codename=codename, architecture="amd64"
                ),
            )
            collection.manager.add_artifact(
                artifact, user=self.playground.get_default_user()
            )
            artifacts.append(artifact)
        resolver = self.resolver()

        lookup = LookupMultiple.model_validate(
            {"collection": "debian@debian:environments"}
        )
        result = resolver.lookup_multiple_artifacts(lookup, label="test")
        self.assertCountEqual(
            result.artifacts,
            [
                DBBaseTaskInputArtifact.from_artifact(artifact)
                for artifact in artifacts
            ],
        )
        with self.assertNumQueries(0):
            self.assertCountEqual(
                [r.artifact for r in result.artifacts], artifacts
            )

        self.assertEqual(result.label, "test")
        self.assertEqual(result.lookup, lookup)

        lookup = LookupMultiple.model_validate({"collection": "debian"})
        result = resolver.lookup_multiple_artifacts(
            lookup,
            default_category=CollectionCategory.ENVIRONMENTS,
            label="test",
        )
        self.assertCountEqual(
            result.artifacts,
            [
                DBBaseTaskInputArtifact.from_artifact(artifact)
                for artifact in artifacts
            ],
        )
        with self.assertNumQueries(0):
            self.assertCountEqual(
                [r.artifact for r in result.artifacts], artifacts
            )
        self.assertEqual(result.label, "test")
        self.assertEqual(result.lookup, lookup)

        with self.assertRaisesRegex(
            LookupError,
            "'debian' does not specify a category and the context does not "
            "supply a default",
        ):
            resolver.lookup_multiple_artifacts(
                LookupMultiple.model_validate({"collection": "debian"}),
                label="test",
            )

    def test_find_related_artifacts(self) -> None:
        """Find artifacts via relations."""
        upload_artifacts = self.playground.create_upload_artifacts(
            binaries=[("libhello1", "amd64"), ("hello", "all")]
        )
        resolver = self.resolver()
        empty_lookup_multiple = LookupMultiple(cast(tuple[str], []))

        self.assertCountEqual(
            resolver.find_related_artifacts(
                [upload_artifacts.upload.id],
                target_category=ArtifactCategory.SOURCE_PACKAGE,
                label="test",
            ).artifacts,
            [],
        )

        result = resolver.find_related_artifacts(
            [upload_artifacts.upload.id],
            target_category=ArtifactCategory.SOURCE_PACKAGE,
            relation_type=RelationType.EXTENDS,
            label="test",
        )
        self.assertCountEqual(
            result.artifacts,
            [DBBaseTaskInputArtifact.from_artifact(upload_artifacts.source)],
        )
        with self.assertNumQueries(0):
            self.assertEqual(
                result.artifacts[0].artifact, upload_artifacts.source
            )
        self.assertEqual(result.lookup, empty_lookup_multiple)
        self.assertEqual(result.label, "test")

        result = resolver.find_related_artifacts(
            [upload_artifacts.upload.id],
            target_category=ArtifactCategory.BINARY_PACKAGE,
            label="test",
        )
        self.assertCountEqual(result.artifacts, [])
        self.assertEqual(result.lookup, empty_lookup_multiple)
        self.assertEqual(result.label, "test")

        result = resolver.find_related_artifacts(
            [upload_artifacts.upload.id],
            target_category=ArtifactCategory.BINARY_PACKAGE,
            relation_type=RelationType.EXTENDS,
            label="test",
        )
        self.assertCountEqual(
            result.artifacts,
            [
                DBBaseTaskInputArtifact.from_artifact(artifact)
                for artifact in upload_artifacts.binaries
            ],
        )
        with self.assertNumQueries(0):
            self.assertCountEqual(
                [r.artifact for r in result.artifacts],
                upload_artifacts.binaries,
            )
        self.assertEqual(result.lookup, empty_lookup_multiple)
        self.assertEqual(result.label, "test")

        self.playground.create_artifact_relation(
            artifact=upload_artifacts.upload,
            target=upload_artifacts.source,
            relation_type=ArtifactRelation.Relations.RELATES_TO,
        )
        result = resolver.find_related_artifacts(
            [upload_artifacts.upload.id],
            target_category=ArtifactCategory.SOURCE_PACKAGE,
            label="test",
        )
        self.assertCountEqual(
            result.artifacts,
            [DBBaseTaskInputArtifact.from_artifact(upload_artifacts.source)],
        )
        with self.assertNumQueries(0):
            self.assertEqual(
                result.artifacts[0].artifact, upload_artifacts.source
            )
        self.assertEqual(result.lookup, empty_lookup_multiple)
        self.assertEqual(result.label, "test")

    def test_get_fallback_architecture(self) -> None:
        resolver = self.resolver()
        with mock.patch(
            "debusine.db.models.WorkRequest.get_fallback_architecture",
            return_value=mock.sentinel.arch,
        ):
            self.assertEqual(
                resolver.get_fallback_architecture(), mock.sentinel.arch
            )

    def test_export_suite_signing_keys_none(self) -> None:
        collection = self.playground.create_collection(
            "trixie",
            CollectionCategory.SUITE,
            workspace=self.scenario.workspace,
        )
        resolver = self.resolver()

        self.assertIsNone(resolver.export_suite_signing_keys(collection.id))

    def test_export_suite_signing_keys(self) -> None:
        keys = [
            PGPKey.new(PubKeyAlgorithm.EdDSA, EllipticCurveOID.Ed25519).pubkey
            for _ in range(2)
        ]
        for key in keys:
            self.playground.create_asset_usage(
                resource=self.playground.create_signing_key_asset(
                    purpose=KeyPurpose.OPENPGP_REPOSITORY,
                    fingerprint=key.fingerprint,
                    public_key=str(key),
                    workspace=self.scenario.workspace,
                ),
                workspace=self.scenario.workspace,
            )
        collection = self.playground.create_collection(
            "trixie",
            CollectionCategory.SUITE,
            workspace=self.scenario.workspace,
            data={"signing_keys": [key.fingerprint for key in keys]},
        )
        resolver = self.resolver()

        signing_keys = resolver.export_suite_signing_keys(collection.id)
        assert signing_keys is not None
        self.assertTrue(
            signing_keys.startswith("-----BEGIN PGP PUBLIC KEY BLOCK-----\n")
        )
        self.assertEqual(
            PGPKeyring(signing_keys).fingerprints(),
            {key.fingerprint for key in keys},
        )


class EnvironmentInputTests(TestCase):
    """Test EnvironmentInput and OptionalEnvironmentInput field resolution."""

    @override
    def setUp(self) -> None:
        super().setUp()
        # Set up the mock task
        self.task = mock.MagicMock(spec=DBWorkerProxyTask)
        self.task.name = mock.sentinel.task_name
        self.task.data = mock.MagicMock()
        self.task.work_request = mock.MagicMock(spec=WorkRequest)
        self.task.task = mock.MagicMock(spec=BaseTaskWithExecutor)
        self.task.task.name = mock.sentinel.task_name
        self.task.task.backend = mock.sentinel.backend
        # Fields are resolved in order, and this needs to have been already
        # resolved
        self.task.task.build_architecture = mock.sentinel.build_architecture

        # Set up the resolver
        self.resolver = DBFieldResolver(self.task)
        self.resolver_lookup_environment = self.enterContext(
            mock.patch.object(
                self.resolver,
                "lookup_environment",
                return_value=mock.sentinel.environment,
            )
        )
        self.resolver_get_field = self.enterContext(
            mock.patch.object(
                self.resolver, "get_field", return_value=mock.sentinel.lookup
            )
        )

    def setup[F: inputs.DataFieldInput[Any, Any]](self, field: F) -> F:
        """Set up the field to test."""
        field.field = mock.sentinel.field_name
        self.field_check_category = self.enterContext(
            mock.patch.object(field, "check_category")
        )
        return field

    def assertResolves[F](
        self,
        method: Callable[[F], Any],
        field: F,
        *,
        backend: BackendType | None = mock.sentinel.backend,
        image_category: ExecutorImageCategory | None = None,
        try_variant: str | None = mock.sentinel.task_name,
    ) -> None:
        result = method(field)
        self.assertIs(result, mock.sentinel.environment)
        self.field_check_category.assert_called_once_with(
            mock.sentinel.environment
        )
        self.resolver_lookup_environment.assert_called_once_with(
            mock.sentinel.lookup,
            architecture=mock.sentinel.build_architecture,
            backend=backend,
            default_category=CollectionCategory.ENVIRONMENTS,
            image_category=image_category,
            try_variant=try_variant,
            label=mock.sentinel.field_name,
        )

    def test_nonopt_defaults(self) -> None:
        field = self.setup(inputs.EnvironmentInput(field="field"))
        self.assertResolves(self.resolver.resolve_environmentinput, field)

    def test_nonopt_set_backend_false(self) -> None:
        field = self.setup(
            inputs.EnvironmentInput(set_backend=False, field="field")
        )
        self.assertResolves(
            self.resolver.resolve_environmentinput, field, backend=None
        )

    def test_nonopt_try_variant_false(self) -> None:
        field = self.setup(
            inputs.EnvironmentInput(try_variant=False, field="field")
        )
        self.assertResolves(
            self.resolver.resolve_environmentinput, field, try_variant=None
        )

    def test_nonopt_image_category(self) -> None:
        field = self.setup(
            inputs.EnvironmentInput(
                image_category=ExecutorImageCategory.TARBALL, field="field"
            )
        )
        self.assertResolves(
            self.resolver.resolve_environmentinput,
            field,
            image_category=ExecutorImageCategory.TARBALL,
        )

    def test_nonopt_field_is_required(self) -> None:
        field = self.setup(inputs.EnvironmentInput(field="field"))
        self.resolver_get_field.return_value = None
        with self.assertRaises(AssertionError):
            self.resolver.resolve_environmentinput(field)

    def test_opt_defaults(self) -> None:
        field = self.setup(inputs.OptionalEnvironmentInput(field="field"))
        self.assertResolves(
            self.resolver.resolve_optionalenvironmentinput, field
        )

    def test_opt_missing(self) -> None:
        field = self.setup(inputs.OptionalEnvironmentInput(field="field"))
        self.resolver_get_field.return_value = None
        self.assertIsNone(self.resolver.resolve_optionalenvironmentinput(field))
        self.field_check_category.assert_not_called()
        self.resolver_lookup_environment.assert_not_called()


class SuiteArchiveInputTests(TestCase):
    """Test SuiteArchiveInput field resolution."""

    scenario = scenarios.DefaultContext()

    work_request: ClassVar[WorkRequest]
    sid: ClassVar[Collection]
    archive: ClassVar[Collection]

    @classmethod
    @override
    def setUpTestData(cls) -> None:
        """Set up common data for tests."""
        super().setUpTestData()
        cls.work_request = cls.playground.create_worker_task()
        cls.sid = cls.playground.create_collection(
            name="sid", category=CollectionCategory.SUITE
        )
        cls.archive = cls.playground.create_singleton_collection(
            category=CollectionCategory.ARCHIVE
        )

    @override
    def setUp(self) -> None:
        super().setUp()
        self.task = self.work_request.get_task()

        # Set up the resolver
        self.resolver = DBFieldResolver(self.task)
        self.resolver_get_field = self.enterContext(
            mock.patch.object(
                self.resolver, "get_field", return_value="sid@debian:suite"
            )
        )

        # Set up the Field
        self.field = inputs.SuiteArchiveInput(field="field_name")

        self.wr_lookup_single = self.enterContext(
            self.spy("debusine.db.models.WorkRequest.lookup_single")
        )
        self.lookup_single = self.enterContext(
            self.spy("debusine.server.collections.lookup.lookup_single")
        )

    def test_suite_only(self) -> None:
        self.archive.delete()
        suite, archive = self.resolver.resolve_suitearchiveinput(self.field)
        self.assertEqual(suite.collection_id, self.sid.id)
        self.assertIsNone(archive)
        self.wr_lookup_single.assert_has_calls(
            [
                mock.call(
                    self.work_request,
                    lookup=self.resolver_get_field.return_value,
                    expect_type=LookupChildType.COLLECTION,
                ),
            ]
        )
        self.lookup_single.assert_has_calls(
            [
                mock.call(
                    "_@debian:archive",
                    workspace=self.scenario.workspace,
                    pc=PermissionContext.for_work_request(
                        work_request=self.work_request
                    ),
                    expect_type=LookupChildType.COLLECTION,
                ),
            ]
        )

    def test_suite_archive(self) -> None:
        suite, archive = self.resolver.resolve_suitearchiveinput(self.field)
        self.assertEqual(suite.collection_id, self.sid.id)
        assert archive is not None
        self.assertEqual(archive.collection_id, self.archive.id)
        self.wr_lookup_single.assert_has_calls(
            [
                mock.call(
                    self.work_request,
                    lookup=self.resolver_get_field.return_value,
                    expect_type=LookupChildType.COLLECTION,
                ),
            ]
        )
        self.lookup_single.assert_has_calls(
            [
                mock.call(
                    "_@debian:archive",
                    workspace=self.scenario.workspace,
                    pc=PermissionContext.for_work_request(
                        work_request=self.work_request
                    ),
                    expect_type=LookupChildType.COLLECTION,
                ),
            ]
        )

    def test_suite_archive_have_cached_models(self) -> None:
        suite, archive = self.resolver.resolve_suitearchiveinput(self.field)
        assert archive is not None
        with self.assertNumQueries(0):
            self.assertEqual(suite.collection, self.sid)
        with self.assertNumQueries(0):
            self.assertEqual(archive.collection, self.archive)

    def test_lookup_archive_in_suite_workspace(self) -> None:
        ws_other = self.playground.create_workspace(name="other", public=True)
        sid_other = self.playground.create_collection(
            name="sid", category=CollectionCategory.SUITE, workspace=ws_other
        )
        archive_other = self.playground.create_singleton_collection(
            category=CollectionCategory.ARCHIVE, workspace=ws_other
        )

        self.resolver_get_field.return_value = f"{sid_other.id}@collections"

        suite, archive = self.resolver.resolve_suitearchiveinput(self.field)
        self.assertEqual(suite.collection_id, sid_other.id)
        assert archive is not None
        self.assertEqual(archive.collection_id, archive_other.id)
        self.wr_lookup_single.assert_has_calls(
            [
                mock.call(
                    self.work_request,
                    lookup=self.resolver_get_field.return_value,
                    expect_type=LookupChildType.COLLECTION,
                ),
            ]
        )
        self.lookup_single.assert_has_calls(
            [
                mock.call(
                    "_@debian:archive",
                    workspace=ws_other,
                    pc=PermissionContext.for_work_request(
                        work_request=self.work_request
                    ),
                    expect_type=LookupChildType.COLLECTION,
                ),
            ]
        )

    def test_wrong_suite_category(self) -> None:
        self.sid.category = CollectionCategory.TEST
        self.sid.save()
        self.resolver_get_field.return_value = "sid@debusine:test"
        with self.assertRaisesRegex(
            TaskConfigError,
            rf"^field_name: unexpected collection category: "
            rf"'{CollectionCategory.TEST}'\. "
            rf"Expected: '{CollectionCategory.SUITE}'$",
        ):
            self.resolver.resolve_suitearchiveinput(self.field)


@override_settings(DEBUSINE_DEBIAN_ARCHIVE_PRIMARY_FQDN="deb.example.com")
class ExtraRepositoriesInputTests(TestCase):
    """Test ExtraRepositoriesInput field resolution."""

    scenario = scenarios.DefaultContext()
    work_request: ClassVar[WorkRequest]
    sid: ClassVar[Collection]
    keys: ClassVar[list[PGPKey]]

    @classmethod
    @override
    def setUpTestData(cls) -> None:
        """Set up common data for tests."""
        super().setUpTestData()
        cls.work_request = cls.playground.create_worker_task()

        cls.keys = [
            PGPKey.new(PubKeyAlgorithm.EdDSA, EllipticCurveOID.Ed25519).pubkey
            for _ in range(2)
        ]
        for key in cls.keys:
            cls.playground.create_asset_usage(
                resource=cls.playground.create_signing_key_asset(
                    purpose=KeyPurpose.OPENPGP_REPOSITORY,
                    fingerprint=key.fingerprint,
                    public_key=str(key),
                    workspace=cls.scenario.workspace,
                ),
                workspace=cls.scenario.workspace,
            )
        cls.sid = cls.playground.create_collection(
            name="sid",
            category=CollectionCategory.SUITE,
            workspace=cls.scenario.workspace,
            data={
                "components": ["main"],
                "signing_keys": [cls.keys[0].fingerprint],
            },
        )

    @override
    def setUp(self) -> None:
        super().setUp()
        self.task = self.work_request.get_task()

        # Set up the resolver
        self.resolver = DBFieldResolver(self.task)
        self.resolver_get_field = self.enterContext(
            mock.patch.object(
                self.resolver, "get_field", return_value=mock.sentinel.lookup
            )
        )

        # Set up the Field
        self.field = inputs.ExtraRepositoriesInput(field="field_name")

        self.lookup_single = self.enterContext(
            self.spy("debusine.db.models.WorkRequest.lookup_single")
        )

    def test_resolve_none(self) -> None:
        self.resolver_get_field.return_value = None
        resolve_extra_repository = self.enterContext(
            mock.patch.object(
                self.resolver,
                "_resolve_extra_repository",
                return_value=mock.sentinel.resolved,
            )
        )
        self.assertIsNone(
            self.resolver.resolve_extrarepositoriesinput(self.field)
        )
        resolve_extra_repository.assert_not_called()

    def test_resolve_empty_list(self) -> None:
        self.resolver_get_field.return_value = []
        resolve_extra_repository = self.enterContext(
            mock.patch.object(
                self.resolver,
                "_resolve_extra_repository",
                return_value=mock.sentinel.resolved,
            )
        )
        self.assertEqual(
            self.resolver.resolve_extrarepositoriesinput(self.field), []
        )
        resolve_extra_repository.assert_not_called()

    def test_resolve_list_items(self) -> None:
        self.resolver_get_field.return_value = [
            mock.sentinel.first,
            mock.sentinel.second,
        ]
        resolve_extra_repository = self.enterContext(
            mock.patch.object(
                self.resolver,
                "_resolve_extra_repository",
                return_value=mock.sentinel.resolved,
            )
        )
        self.assertEqual(
            self.resolver.resolve_extrarepositoriesinput(self.field),
            [mock.sentinel.resolved, mock.sentinel.resolved],
        )
        resolve_extra_repository.assert_has_calls(
            [
                mock.call(mock.sentinel.first),
                mock.call(mock.sentinel.second),
            ]
        )

    def test_resolve_extra_repo_external(self) -> None:
        external = ExtraExternalRepository(
            url="https://example.org", suite="sid", components=["main"]
        )
        self.assertIs(
            self.resolver._resolve_extra_repository(external),
            external,
        )

    def test_debusine(self) -> None:
        debusine = ExtraDebusineRepository(
            suite="sid@debian:suite", components=["main"]
        )
        resolved = self.resolver._resolve_extra_repository(debusine)
        self.assertEqual(
            resolved,
            ExtraExternalRepository(
                url="http://deb.example.com/debusine/System",
                suite="sid",
                components=["main"],
                signing_key=str(ExportablePGPKeyring(self.keys[0])),
            ),
        )

    def test_missing_components(self) -> None:
        del self.sid.data["components"]
        self.sid.save()

        debusine = ExtraDebusineRepository(
            suite="sid@debian:suite", components=None
        )
        with self.assertRaisesRegex(
            TaskConfigError,
            "'components' not set for sid@debian:suite",
        ):
            self.resolver._resolve_extra_repository(debusine)

    def test_default_components(self) -> None:
        self.sid.data["components"] = ["main", "contrib", "non-free"]
        self.sid.save()

        debusine = ExtraDebusineRepository(
            suite="sid@debian:suite", components=None
        )
        resolved = self.resolver._resolve_extra_repository(debusine)
        self.assertEqual(
            resolved,
            ExtraExternalRepository(
                url="http://deb.example.com/debusine/System",
                suite="sid",
                components=["main", "contrib", "non-free"],
                signing_key=str(ExportablePGPKeyring(self.keys[0])),
            ),
        )

    def test_explicit_components(self) -> None:
        self.sid.data["components"] = ["main", "contrib", "non-free"]
        self.sid.save()

        debusine = ExtraDebusineRepository(
            suite="sid@debian:suite", components=["main", "contrib"]
        )
        resolved = self.resolver._resolve_extra_repository(debusine)
        self.assertEqual(
            resolved,
            ExtraExternalRepository(
                url="http://deb.example.com/debusine/System",
                suite="sid",
                components=["main", "contrib"],
                signing_key=str(ExportablePGPKeyring(self.keys[0])),
            ),
        )


class ConcreteDataFieldInput(inputs.DataFieldInput[int, None]):
    """Concrete version of DataFieldInput."""

    @override
    def resolve(self, field_resolver: FieldResolver) -> None:
        raise NotImplementedError()


class GetFieldTests(TestCase):
    """Test :py:func:`get_field`."""

    @override
    def setUp(self) -> None:
        super().setUp()
        # Set up the mock task
        self.task = mock.MagicMock()
        self.task.data = SimpleNamespace()
        self.task.work_request = mock.MagicMock(spec=WorkRequest)
        self.resolver = DBFieldResolver(self.task)

    def test_get_field(self) -> None:
        self.task.data.value = 1
        self.task.data.unset = None
        self.task.data.nested = SimpleNamespace()
        self.task.data.nested.value = 2

        for field, expected in (
            ("value", 1),
            ("nested.value", 2),
            ("unset", None),
        ):
            with self.subTest(field=field):
                f = ConcreteDataFieldInput(field=field)
                self.assertEqual(self.resolver.get_field(f), expected)

    def test_get_field_fails(self) -> None:
        for field in ("value", "nested.value"):
            with (
                self.subTest(Field=field),
                self.assertRaisesRegex(
                    TaskConfigError,
                    re.escape(f"Task input attribute {field!r} not found"),
                ),
            ):
                f = ConcreteDataFieldInput(field=field)
                self.resolver.get_field(f)

    def test_get_field_invalid_type(self) -> None:
        f = ConcreteDataFieldInput(field="value")
        value: Any
        for value in ("foo", [], {}):
            with self.subTest(value=repr(value)):
                self.task.data.value = value
                with self.assertRaisesRegex(
                    TaskConfigError,
                    re.escape(
                        f"Task input attribute 'value' is {value!r}"
                        f" which is not of type <class 'int'>"
                    ),
                ):
                    self.resolver.get_field(f)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for classes in models.py."""

import re
from contextlib import ExitStack
from typing import Any
from unittest import TestCase

import pydantic

import debusine.artifacts.models as data_models
from debusine.assets.models import KeyPurpose


class EnumTests(TestCase):
    """Tests for model enums."""

    def test_enums_str(self) -> None:
        """Test enum stringification."""
        for enum_cls in (
            data_models.DebianAutopkgtestResultStatus,
            data_models.DebianLintianSeverity,
        ):
            with self.subTest(enum_cls=repr(enum_cls)):
                for el in enum_cls:
                    with self.subTest(el=el.value):
                        self.assertEqual(str(el), el.value)


class DebusineHistoricalTaskRunTests(TestCase):
    """Test the DebusineHistoricalTaskRun model."""

    def test_task_type_validation(self) -> None:
        """task_type must be a valid task type."""
        kwargs: dict[str, Any] = {
            "task_name": "noop",
            "subject": "subject",
            "context": "sid/amd64",
            "timestamp": 0,
            "work_request_id": 1,
            "result": "success",
            "runtime_statistics": data_models.RuntimeStatistics(duration=1),
        }
        data_models.DebusineHistoricalTaskRun(
            task_type=data_models.TaskTypes.WORKER, **kwargs
        )
        with self.assertRaisesRegex(ValueError, r"Input should be.*type=enum"):
            data_models.DebusineHistoricalTaskRun(
                task_type="does-not-exist", **kwargs
            )


class DebusineTaskConfigurationTests(TestCase):
    """Test the DebusineTaskConfiguration model."""

    def test_name(self) -> None:
        """Test name generation."""
        data = data_models.DebusineTaskConfiguration(template="test")
        self.assertEqual(data.name(), "template:test")

        data = data_models.DebusineTaskConfiguration(
            task_type=data_models.TaskTypes.WORKER,
            task_name="noop",
            subject="subject",
            context="sid/amd64",
        )
        self.assertEqual(data.name(), "Worker:noop:subject:sid/amd64")

        data = data_models.DebusineTaskConfiguration(
            task_type=data_models.TaskTypes.WORKER,
            task_name="noop",
            subject="subject",
            context="source:debusine",
        )
        self.assertEqual(data.name(), "Worker:noop:subject:source%3Adebusine")

    def test_task_type_validation(self) -> None:
        """task_type must be a valid task type."""
        kwargs: dict[str, Any] = {
            "task_name": "noop",
            "subject": "subject",
            "context": "sid/amd64",
        }
        data_models.DebusineTaskConfiguration(
            task_type=data_models.TaskTypes.WORKER, **kwargs
        )
        with self.assertRaisesRegex(ValueError, r"Input should be.*type=enum"):
            data_models.DebusineTaskConfiguration(
                task_type="does-not-exist", **kwargs
            )
        with self.assertRaisesRegex(ValueError, "should be set"):
            data_models.DebusineTaskConfiguration(task_type=None, **kwargs)
        with self.assertRaisesRegex(ValueError, "should be set"):
            data_models.DebusineTaskConfiguration(**kwargs)

    def test_task_name_validation(self) -> None:
        """Task names must not contain colons."""
        kwargs: dict[str, Any] = {
            "task_type": data_models.TaskTypes.WORKER,
            "subject": "subject",
            "context": "sid/amd64",
        }
        data_models.DebusineTaskConfiguration(task_name="noop", **kwargs)
        with self.assertRaisesRegex(ValueError, "may not contain ':'"):
            data_models.DebusineTaskConfiguration(task_name="no:op", **kwargs)
        with self.assertRaisesRegex(ValueError, "should be set"):
            data_models.DebusineTaskConfiguration(task_name=None, **kwargs)
        with self.assertRaisesRegex(ValueError, "should be set"):
            data_models.DebusineTaskConfiguration(**kwargs)

    def test_subject_validation(self) -> None:
        """Subject must not contain colons."""
        kwargs: dict[str, Any] = {
            "task_type": data_models.TaskTypes.WORKER,
            "task_name": "noop",
            "context": "sid/amd64",
        }
        data_models.DebusineTaskConfiguration(subject="subject", **kwargs)
        data_models.DebusineTaskConfiguration(subject=None, **kwargs)
        data_models.DebusineTaskConfiguration(**kwargs)
        with self.assertRaisesRegex(ValueError, "may not contain ':'"):
            data_models.DebusineTaskConfiguration(subject="sub:ject", **kwargs)

    def test_context_validation(self) -> None:
        """Context must not contain colons."""
        kwargs: dict[str, Any] = {
            "task_type": data_models.TaskTypes.WORKER,
            "task_name": "noop",
            "subject": "subject",
        }
        data_models.DebusineTaskConfiguration(
            context="source:debusine", **kwargs
        )
        data_models.DebusineTaskConfiguration(context="sid/amd64", **kwargs)
        data_models.DebusineTaskConfiguration(context="a b c", **kwargs)
        data_models.DebusineTaskConfiguration(context=None, **kwargs)
        data_models.DebusineTaskConfiguration(**kwargs)

    def test_template_excludes_task_match(self) -> None:
        """Template excludes the use of task match fields."""
        for name, value in (
            ("task_type", data_models.TaskTypes.WORKER),
            ("task_name", "noop"),
            ("subject", "subject"),
            ("context", "sid/amd64"),
        ):
            with (
                self.subTest(name=name),
                self.assertRaisesRegex(
                    ValueError, "should be empty if template is set"
                ),
            ):
                kwargs: dict[str, Any] = {name: value}
                data_models.DebusineTaskConfiguration(template="test", **kwargs)

    def test_get_lookup_names(self) -> None:
        """Test get_lookup_names."""
        f = data_models.DebusineTaskConfiguration.get_lookup_names
        for args, expected in (
            (("t", "n", None, None), ["t:n::"]),
            (("t", "n", "s", None), ["t:n::", "t:n:s:"]),
            (("t", "n", None, "c"), ["t:n::", "t:n::c"]),
            (("t", "n", "s", "c"), ["t:n::", "t:n::c", "t:n:s:", "t:n:s:c"]),
            (
                ("t", "n", "s", ":"),
                ["t:n::", "t:n::%3A", "t:n:s:", "t:n:s:%3A"],
            ),
        ):
            with self.subTest(args=args):
                self.assertEqual(f(*args), expected)

    def test_path_validation_succeeds(self) -> None:
        for path in (
            None,
            "foo.yaml",
            "foo/bar.yaml",
            "foo..bar.yaml",
            ".foo/bar.yaml",
            "..foo.yaml",
            "foo...yaml",
        ):
            with self.subTest(path=path):
                item = data_models.DebusineTaskConfiguration(
                    template="test", path=path
                )
                self.assertEqual(item.path, path)

    def test_path_validation_fails(self) -> None:
        absolute = "must not be an absolute path"
        not_yaml = r"must end with \.yaml or \.yml"
        single_dot = r"must not contain '\.' as a path element"
        double_dot = r"must not contain '\.\.' as a path element"

        for path, msg in (
            ("", not_yaml),
            ("/", not_yaml),
            ("/foo", not_yaml),
            ("/foo.yaml", absolute),
            ("./", not_yaml),
            ("./foo.yaml", single_dot),
            ("foo/./bar.yaml", single_dot),
            ("foo/.", not_yaml),
            ("foo.yaml/", not_yaml),
            ("../foo.yaml", double_dot),
            ("foo/bar/../baz.yaml", double_dot),
            ("foo/bar/baz/..", not_yaml),
        ):
            with (
                self.subTest(path=path),
                self.assertRaisesRegex(
                    ValueError, rf"{re.escape(repr(path))} {msg}"
                ),
            ):
                data_models.DebusineTaskConfiguration(
                    template="test", path=path
                )

    def test_path_position_length(self) -> None:
        for path, pos, validates in (
            ("x.yaml", 0, True),
            ("x.yaml", 12345678901234567890, True),
            ("x" * 248 + ".yaml", 0, True),
            ("x" * 249 + ".yaml", 1, False),
            ("x" * 248 + ".yaml", 10, False),
            ("x" * 245 + ".yaml", 1000, True),
            ("x" * 245 + ".yaml", 10000, False),
        ):
            with self.subTest(path=path, pos=pos), ExitStack() as stack:
                if not validates:
                    stack.enter_context(
                        self.assertRaisesRegex(
                            ValueError,
                            r"path:position string .+ characters"
                            " and must be no longer than 255",
                        )
                    )
                data_models.DebusineTaskConfiguration(
                    template="test", path=path, position=pos
                )

    def test_taglist_tagexpr_validation(self) -> None:
        for value, valid in (
            ([], True),
            (["a"], True),
            ("", False),
            ("a", True),
            ("a or b", True),
            ("a or", False),
        ):
            with self.subTest(value=value), ExitStack() as stack:
                if not valid:
                    stack.enter_context(
                        self.assertRaisesRegex(
                            ValueError, "invalid tag expression:"
                        )
                    )
                data_models.DebusineTaskConfiguration(
                    template="test", tags_provided=value
                )
                data_models.DebusineTaskConfiguration(
                    template="test", tags_required=value
                )


class EmptyArtifactDataTests(TestCase):
    """Test the EmptyArtifactData model."""

    def test_get_label(self) -> None:
        """get_label returns None."""
        data = data_models.EmptyArtifactData()
        label = data.get_label()
        self.assertIsNone(label)


class DebianPackageBuildLogTests(TestCase):
    """Test the DebianPackageBuildLog model."""

    def test_get_label(self) -> None:
        """get_label returns source."""
        data = data_models.DebianPackageBuildLog(
            source="test",
            version="1.0-1",
            architecture="amd64",
            filename="test_1.0-1_amd64.buildlog",
        )
        self.assertEqual(data.get_label(), "test_1.0-1_amd64.buildlog")


class DebianSourcePackageTests(TestCase):
    """Test the DebianSourcePackage model."""

    def test_construct(self) -> None:
        """Test model constructor."""
        data = data_models.DebianSourcePackage(
            name="test-name",
            version="test-version",
            type="dpkg",
            dsc_fields={"key": "val"},
        )

        self.assertEqual(data.name, "test-name")
        self.assertEqual(data.version, "test-version")
        self.assertEqual(data.type, "dpkg")
        self.assertEqual(data.dsc_fields, {"key": "val"})
        self.assertEqual(data.get_label(), "test-name_test-version")

    def test_dsc_fields_missing(self) -> None:
        """Test checking that dsc_fields is present."""
        error_msg = r"dsc_fields\s+Field required.*type=missing"

        with self.assertRaisesRegex(Exception, error_msg):
            data_models.DebianSourcePackage(  # type: ignore[call-arg]
                name="test-name", version="test-version", type="dpkg"
            )


class DebianUploadTests(TestCase):
    """Test the DebianUpload model."""

    def test_label(self) -> None:
        """Test get_label scenarios."""

        def make_upload(
            files: list[str], **kwargs: Any
        ) -> data_models.DebianUpload:
            kwargs["Files"] = [{"name": name} for name in files]
            return data_models.DebianUpload(
                type="dpkg",
                changes_fields={"Architecture": "amd64", **kwargs},
            )

        self.assertEqual(
            make_upload(
                ["test.deb"], Source="test", Version="1.0-1"
            ).get_label(),
            "test_1.0-1",
        )
        self.assertEqual(
            make_upload(["test.deb"], Source="test").get_label(),
            "test.deb",
        )
        self.assertEqual(
            make_upload(["test.deb", "test.changes"]).get_label(),
            "test.changes",
        )
        self.assertEqual(
            make_upload(["test2.deb", "test1.deb"]).get_label(),
            "test2.deb",
        )
        upload = make_upload(["test.deb"])
        upload.changes_fields["Files"] = []
        self.assertIsNone(upload.get_label())

    def test_metadata_contains_architecture(self) -> None:
        """changes_fields must contain Architecture."""
        data = {
            "Files": [{"name": "foo.dsc"}],
        }
        with self.assertRaisesRegex(
            ValueError, r"changes_fields must contain Architecture"
        ):
            data_models.DebianUpload(type="dpkg", changes_fields=data)

    def test_metadata_contains_files(self) -> None:
        """changes_fields must contain Architecture."""
        data = {
            "Architecture": "amd64",
        }
        with self.assertRaisesRegex(
            ValueError, r"changes_fields must contain Files"
        ):
            data_models.DebianUpload(type="dpkg", changes_fields=data)

    def test_metadata_contains_debs_if_binary_requires_debs_for_binaries(
        self,
    ) -> None:
        """metadata_contains_debs_if_binary requires at least one .deb."""
        data = {
            "Architecture": "amd64",
            "Files": [{"name": "foo.dsc"}],
        }
        with self.assertRaisesRegex(
            ValueError,
            r"No \.debs found in \['foo\.dsc'\] which is expected to contain "
            r"binaries for amd64",
        ):
            data_models.DebianUpload.metadata_contains_debs_if_binary(data)

    def test_metadata_contains_debs_if_binary_ignores_source_uploads(
        self,
    ) -> None:
        """metadata_contains_debs_if_binary ignores source uploads."""
        data = {
            "Architecture": "source",
            "Files": [{"name": "foo.dsc"}],
        }
        self.assertEqual(
            data_models.DebianUpload.metadata_contains_debs_if_binary(data),
            data,
        )

    def test_metadata_contains_debs_if_binary_finds_debs_in_source_uploads(
        self,
    ) -> None:
        """metadata_contains_debs_if_binary finds debs in source uploads."""
        data = {
            "Architecture": "source",
            "Files": [{"name": "foo.deb"}],
        }
        with self.assertRaisesRegex(
            ValueError,
            r"Unexpected binary packages \['foo\.deb'\] found in source-only "
            r"upload\.",
        ):
            data_models.DebianUpload.metadata_contains_debs_if_binary(data)

    def test_metadata_contains_debs_if_binary_accepts_debs(self) -> None:
        """metadata_contains_debs_if_binary will accept one .deb."""
        data = {
            "Architecture": "amd64",
            "Files": [{"name": "foo_amd64.deb"}],
        }
        self.assertEqual(
            data,
            data_models.DebianUpload.metadata_contains_debs_if_binary(data),
        )

    def test_metadata_contains_debs_if_binary_accepts_debs_in_mixed(
        self,
    ) -> None:
        """metadata_contains_debs_if_binary will accept deb in mixed upload."""
        data = {
            "Architecture": "amd64 source",
            "Files": [{"name": "foo_amd64.deb"}],
        }
        self.assertEqual(
            data,
            data_models.DebianUpload.metadata_contains_debs_if_binary(data),
        )

    def test_metadata_contains_dsc_if_source_requires_1_dsc_for_source(
        self,
    ) -> None:
        """metadata_contains_dsc_if_source requires 1 .dsc."""
        data = {
            "Architecture": "source",
            "Files": [{"name": "foo.deb"}],
        }
        with self.assertRaisesRegex(
            ValueError,
            r"Expected to find one and only one source package in source "
            r"upload\. Found \[\].",
        ):
            data_models.DebianUpload.metadata_contains_dsc_if_source(data)

    def test_metadata_contains_dsc_if_source_rejects_2_dsc_for_source(
        self,
    ) -> None:
        """metadata_contains_dsc_if_source rejects 2 .dscs."""
        data = {
            "Architecture": "source",
            "Files": [{"name": "foo.dsc"}, {"name": "bar.dsc"}],
        }
        with self.assertRaisesRegex(
            ValueError,
            r"Expected to find one and only one source package in source "
            r"upload\. Found \['foo\.dsc', 'bar\.dsc'\]\.",
        ):
            data_models.DebianUpload.metadata_contains_dsc_if_source(data)

    def test_metadata_contains_dsc_if_source_ignores_binary_uploads(
        self,
    ) -> None:
        """metadata_contains_dsc_if_source ignores binary uploads."""
        data = {
            "Architecture": "amd64",
            "Files": [{"name": "foo.deb"}],
        }
        self.assertEqual(
            data_models.DebianUpload.metadata_contains_dsc_if_source(data), data
        )

    def test_metadata_contains_dsc_if_source_finds_dsc_in_bin_uploads(
        self,
    ) -> None:
        """metadata_contains_dsc_if_source finds dsc in binary uploads."""
        data = {
            "Architecture": "amd64",
            "Files": [{"name": "foo.dsc"}],
        }
        with self.assertRaisesRegex(
            ValueError,
            r"Binary uploads cannot contain source packages. "
            r"Found: \['foo\.dsc'\].",
        ):
            data_models.DebianUpload.metadata_contains_dsc_if_source(data)

    def test_metadata_contains_dsc_if_source_accepts_1_dsc(self) -> None:
        """metadata_contains_dsc_if_source will accept one .dsc."""
        data = {
            "Architecture": "source",
            "Files": [{"name": "foo.dsc"}],
        }
        self.assertEqual(
            data, data_models.DebianUpload.metadata_contains_dsc_if_source(data)
        )

    def test_metadata_contains_dsc_if_source_accepts_dsc_in_mixed(self) -> None:
        """metadata_contains_dsc_if_source accepts source in mixed upload."""
        data = {
            "Architecture": "amd64 source",
            "Files": [{"name": "foo.dsc"}],
        }
        self.assertEqual(
            data, data_models.DebianUpload.metadata_contains_dsc_if_source(data)
        )


class DebianBinaryPackageTests(TestCase):
    """Test the DebianBinaryPackage model."""

    def test_get_label(self) -> None:
        """get_label returns srcpkg_name."""
        data = data_models.DebianBinaryPackage(
            srcpkg_name="test",
            srcpkg_version="1.0-1",
            deb_fields={
                "Package": "test-bin",
                "Version": "1.0-1+b1",
                "Architecture": "amd64",
            },
            deb_control_files=[],
        )
        self.assertEqual(data.get_label(), "test-bin_1.0-1+b1_amd64")


class DebianBinaryPackagesTests(TestCase):
    """Test the DebianBinaryPackages model."""

    def test_get_label(self) -> None:
        """get_label returns srcpkg_name."""
        data = data_models.DebianBinaryPackages(
            srcpkg_name="test",
            srcpkg_version="1.0-1",
            version="1.0",
            architecture="amd64",
            packages=[],
        )
        self.assertEqual(data.get_label(), "test_1.0-1")


class DebianSystemTarballTests(TestCase):
    """Test the DebianSystemTarball model."""

    def test_get_label(self) -> None:
        """get_label returns filename."""
        data = data_models.DebianSystemTarball(
            filename="bookworm.tar.gz",
            vendor="debian",
            codename="bookworm",
            components=["main"],
            mirror="https://deb.debian.org",
            variant=None,
            pkglist={},
            architecture="amd64",
            with_dev=False,
            with_init=False,
        )
        self.assertEqual(data.get_label(), "bookworm.tar.gz")


class DebianSystemImageTests(TestCase):
    """Test the DebianSystemImage model."""

    def test_get_label(self) -> None:
        """get_label returns filename."""
        data = data_models.DebianSystemImage(
            filename="bookworm.qcow2",
            vendor="debian",
            codename="bookworm",
            components=["main"],
            mirror="https://deb.debian.org",
            variant=None,
            pkglist={},
            architecture="amd64",
            with_dev=False,
            with_init=False,
            image_format="qcow2",
            filesystem="ext4",
            size=123456,
            boot_mechanism="efi",
        )
        self.assertEqual(data.get_label(), "bookworm.qcow2")


class DebianLintianTests(TestCase):
    """Test the DebianLintian model."""

    def test_get_label(self) -> None:
        """get_label returns package names."""

        def make_lintian(**kwargs: str) -> data_models.DebianLintian:
            return data_models.DebianLintian(
                architecture="amd64",
                summary=data_models.DebianLintianSummary(
                    tags_count_by_severity={},
                    package_filename=kwargs,
                    tags_found=[],
                    overridden_tags_found=[],
                    lintian_version="1.0",
                    distribution="bookworm",
                ),
            )

        self.assertEqual(make_lintian().get_label(), "lintian (empty)")
        self.assertEqual(
            make_lintian(hello="hello_1.0-1.dsc").get_label(), "lintian: hello"
        )
        self.assertEqual(
            make_lintian(
                libhello="libhello_1.0-1.deb", hello="hello_1.0-1.deb"
            ).get_label(),
            "lintian: hello, libhello",
        )


class DebianAutopkgtestTests(TestCase):
    """Test the DebianAutopkgtest model."""

    def test_get_label(self) -> None:
        """get_label returns source_package.name."""
        data = data_models.DebianAutopkgtest(
            results={},
            cmdline="",
            source_package=data_models.DebianAutopkgtestSource(
                name="hello",
                version="1.0-1",
                url="https://deb.debian.org/pool/h/hello",
            ),
            architecture="amd64",
            distribution="bookworm",
        )
        self.assertEqual(data.get_label(), "hello")


class DebianPiupartsTests(TestCase):
    """Test the DebianPiuparts model."""

    def test_get_label(self) -> None:
        data = data_models.DebianPiuparts(
            piuparts_version="1.6.0",
            binary_packages={"hello": "1.0", "hello-doc": "1.0"},
            architecture="amd64",
            distribution="debian:bookworm",
        )
        self.assertEqual(data.get_label(), "piuparts: hello, hello-doc")


class DebusineSigningInputTests(TestCase):
    """Test the DebusineSigningInput model."""

    def test_get_label(self) -> None:
        """get_label returns None."""
        data = data_models.DebusineSigningInput(trusted_certs=[])
        label = data.get_label()
        self.assertIsNone(label)


class DebusinePromiseTests(TestCase):
    """Test the DebusinePromise."""

    def test_construct(self) -> None:
        """Test the model constructor with valid data."""
        data = data_models.DebusinePromise(
            promise_work_request_id=1,
            promise_workflow_id=100,
            promise_category="debian:binary-package",
        )

        self.assertEqual(data.promise_work_request_id, 1)
        self.assertEqual(data.promise_workflow_id, 100)
        self.assertEqual(data.promise_category, "debian:binary-package")

    def test_extra_fields_allowed(self) -> None:
        """Test that extra fields are allowed in the model."""
        data = data_models.DebusinePromise(
            promise_work_request_id=1,
            promise_workflow_id=100,
            promise_category="debian:binary-package",
            extra_field="extra_value",  # type: ignore[call-arg]
        )

        self.assertEqual(data.extra_field, "extra_value")  # type: ignore[attr-defined]

    def test_forbidden_promise_prefix_in_extra_field(self) -> None:
        r"""Test that extra fields with 'promise\\_' are not allowed."""
        expected_msg = (
            "Field name 'promise_extra' starting with 'promise_' is not allowed"
        )
        with self.assertRaisesRegex(ValueError, expected_msg):
            data_models.DebusinePromise(
                promise_work_request_id=1,
                promise_workflow_id=100,
                promise_category="debian:binary-package",
                promise_extra="not_allowed",  # type: ignore[call-arg]
            )


class SigningResultTests(TestCase):
    """Test the SigningResult model."""

    def test_output_file(self) -> None:
        """A SigningResult may have an output_file and no error_message."""
        data_models.SigningResult(file="file", output_file="file.sig")

    def test_error_message(self) -> None:
        """A SigningResult may have an error_message and no output_file."""
        data_models.SigningResult(file="file", error_message="Boom")

    def test_both_output_file_and_error_message(self) -> None:
        """A SigningResult may not have both output_file and error_message."""
        with self.assertRaisesRegex(
            ValueError,
            "Exactly one of output_file and error_message must be set",
        ):
            data_models.SigningResult(
                file="file", output_file="file.sig", error_message="Boom"
            )

    def test_neither_output_file_nor_error_message(self) -> None:
        """A SigningResult must have either output_file or error_message."""
        with self.assertRaisesRegex(
            ValueError,
            "Exactly one of output_file and error_message must be set",
        ):
            data_models.SigningResult(file="file")


class DebusineSigningOutputTests(TestCase):
    """Test the DebusineSigningOutput model."""

    def test_get_label(self) -> None:
        """get_label returns None."""
        data = data_models.DebusineSigningOutput(
            purpose=KeyPurpose.UEFI,
            fingerprint="123456ABCD",
            results=[],
        )
        label = data.get_label()
        self.assertIsNone(label)


class DebianBlhcTests(TestCase):
    """Test the DebianBlhc model."""

    def test_get_label_no_architecture(self) -> None:
        data = data_models.DebianBlhc()
        self.assertEqual(data.get_label(), "blhc analysis")

    def test_get_label(self) -> None:
        data = data_models.DebianBlhc(
            source="hello", version="1.0", architecture="amd64"
        )
        self.assertEqual(data.get_label(), "blhc analysis for hello_1.0_amd64")


class DebianDebugSymbolsTests(TestCase):
    """Test the DebianDebugSymbols model."""

    def test_valid_input(self) -> None:
        data = data_models.DebianDebugSymbols(build_ids=["a" * 40, "b" * 40])
        self.assertEqual(data.build_ids, ["a" * 40, "b" * 40])

    def test_invalid_inputs(self) -> None:
        invalid_cases = [
            ("wrong length", ["abc123"], "abc123"),
            ("uppercase", ["A" * 40], "A" * 40),
            ("non hex", ["g" * 40], "g" * 40),
            ("duplicates", ["a" * 40, "a" * 40], "duplicate build_id"),
        ]
        for name, build_ids, expected_msg in invalid_cases:
            with (
                self.subTest(case=name),
                self.assertRaisesRegex(pydantic.ValidationError, expected_msg),
            ):
                data_models.DebianDebugSymbols(build_ids=build_ids)

    def test_get_label_single(self) -> None:
        data = data_models.DebianDebugSymbols(build_ids=["a" * 40])
        self.assertEqual(data.get_label(), "debug-symbols (1 build-id)")

    def test_get_label_double(self) -> None:
        data = data_models.DebianDebugSymbols(build_ids=["a" * 40, "b" * 40])
        self.assertEqual(data.get_label(), "debug-symbols (2 build-ids)")


class DebDiffTests(TestCase):
    """Test the DebDiff model."""

    def test_get_label_override(self) -> None:
        data = data_models.DebDiff(
            original="original-package",
            new="new-package",
            label="A → B",
        )
        self.assertEqual(data.get_label(), "A → B")

    def test_get_label(self) -> None:
        """get_label returns the debdiff command."""
        data = data_models.DebDiff(
            original="original-package",
            new="new-package",
        )
        self.assertEqual(
            data.get_label(), "debdiff original-package new-package"
        )


class DebianRepositoryIndexTests(TestCase):
    """Test the DebianRepositoryIndex model."""

    def test_get_label(self) -> None:
        data = data_models.DebianRepositoryIndex(path="main/source/Sources.xz")
        self.assertEqual(data.get_label(), "main/source/Sources.xz")


class GetSourcePackageNameTests(TestCase):
    """Tests for function get_source_package_name."""

    def test_DebianSourcePackage(self) -> None:
        artifact_data = data_models.DebianSourcePackage(
            name="hello", version="1.0", type="dpkg", dsc_fields={}
        )
        self.assertEqual(
            data_models.get_source_package_name(artifact_data), "hello"
        )

    def test_DebianUpload(self) -> None:
        artifact_data = data_models.DebianUpload(
            type="dpkg",
            changes_fields={
                "Source": "hello",
                "Architecture": "amd64",
                "Files": [{"name": "hello_1.0_amd64.deb"}],
            },
        )
        self.assertEqual(
            data_models.get_source_package_name(artifact_data), "hello"
        )

    def test_DebianBinaryPackage(self) -> None:
        artifact_data = data_models.DebianBinaryPackage(
            srcpkg_name="hello",
            srcpkg_version="1.0-1",
            deb_fields={
                "Package": "hello-bin",
                "Version": "1.0-1+b1",
                "Architecture": "amd64",
            },
            deb_control_files=[],
        )
        self.assertEqual(
            data_models.get_source_package_name(artifact_data), "hello"
        )

    def test_DebianBinaryPackages(self) -> None:
        artifact_data = data_models.DebianBinaryPackages(
            srcpkg_name="hello",
            srcpkg_version="1.0-1",
            version="1.0",
            architecture="amd64",
            packages=[],
        )
        self.assertEqual(
            data_models.get_source_package_name(artifact_data), "hello"
        )

    def test_invalid_type(self) -> None:
        artifact_data = data_models.EmptyArtifactData()

        with self.assertRaisesRegex(
            TypeError, "^Unexpected type: EmptyArtifactData$"
        ):
            data_models.get_source_package_name(
                artifact_data  # type: ignore[arg-type]
            )


class GetSourcePackageVersionTests(TestCase):
    """Tests for function get_source_package_version."""

    def test_DebianSourcePackage(self) -> None:
        artifact_data = data_models.DebianSourcePackage(
            name="hello", version="1.0", type="dpkg", dsc_fields={}
        )
        self.assertEqual(
            data_models.get_source_package_version(artifact_data), "1.0"
        )

    def test_DebianUpload(self) -> None:
        artifact_data = data_models.DebianUpload(
            type="dpkg",
            changes_fields={
                "Source": "hello",
                "Version": "1.0",
                "Architecture": "source",
                "Files": [{"name": "hello_1.0_source.dsc"}],
            },
        )
        self.assertEqual(
            data_models.get_source_package_version(artifact_data), "1.0"
        )

    def test_DebianBinaryPackage(self) -> None:
        artifact_data = data_models.DebianBinaryPackage(
            srcpkg_name="test",
            srcpkg_version="1.0",
            deb_fields={
                "Package": "test-bin",
                "Version": "1.0-1+b1",
                "Architecture": "amd64",
            },
            deb_control_files=[],
        )
        self.assertEqual(artifact_data.get_label(), "test-bin_1.0-1+b1_amd64")

        self.assertEqual(
            data_models.get_source_package_version(artifact_data), "1.0"
        )

    def test_DebianBinaryPackages(self) -> None:
        artifact_data = data_models.DebianBinaryPackages(
            srcpkg_name="test",
            srcpkg_version="1.0",
            version="1.0-1",
            architecture="amd64",
            packages=[],
        )

        self.assertEqual(
            data_models.get_source_package_version(artifact_data), "1.0"
        )

    def test_invalid_type(self) -> None:
        artifact_data = data_models.EmptyArtifactData()

        with self.assertRaisesRegex(
            TypeError, "^Unexpected type: EmptyArtifactData$"
        ):
            data_models.get_source_package_version(
                artifact_data  # type: ignore[arg-type]
            )


class GetBinaryPackageNameTests(TestCase):
    """Tests for function get_binary_package_name."""

    def test_DebianBinaryPackage(self) -> None:
        artifact_data = data_models.DebianBinaryPackage(
            srcpkg_name="hello",
            srcpkg_version="1.0-1",
            deb_fields={
                "Package": "hello-bin",
                "Version": "1.0-1+b1",
                "Architecture": "amd64",
            },
            deb_control_files=[],
        )
        self.assertEqual(
            data_models.get_binary_package_name(artifact_data), "hello-bin"
        )

    def test_invalid_type(self) -> None:
        artifact_data = data_models.EmptyArtifactData()

        with self.assertRaisesRegex(
            TypeError, "^Unexpected type: EmptyArtifactData$"
        ):
            data_models.get_binary_package_name(
                artifact_data  # type: ignore[arg-type]
            )


class GetBinaryPackageVersionTests(TestCase):
    """Tests for function get_binary_package_version."""

    def test_DebianUpload(self) -> None:
        artifact_data = data_models.DebianUpload(
            type="dpkg",
            changes_fields={
                "Source": "hello",
                "Version": "1.0-1",
                "Architecture": "amd64",
                "Files": [{"name": "hello_1.0-1_amd64.deb"}],
            },
        )

        self.assertEqual(
            data_models.get_binary_package_version(artifact_data), "1.0-1"
        )

    def test_DebianBinaryPackage(self) -> None:
        artifact_data = data_models.DebianBinaryPackage(
            srcpkg_name="test",
            srcpkg_version="1.0-1",
            deb_fields={
                "Package": "test-bin",
                "Version": "1.0-1+b1",
                "Architecture": "amd64",
            },
            deb_control_files=[],
        )

        self.assertEqual(
            data_models.get_binary_package_version(artifact_data), "1.0-1+b1"
        )

    def test_invalid_type(self) -> None:
        artifact_data = data_models.EmptyArtifactData()

        with self.assertRaisesRegex(
            TypeError, "^Unexpected type: EmptyArtifactData$"
        ):
            data_models.get_binary_package_version(
                artifact_data  # type: ignore[arg-type]
            )


class GetArchitecturesTests(TestCase):
    """Tests for function get_architectures."""

    def test_DebianBinaryPackages(self) -> None:
        artifact_data = data_models.DebianBinaryPackage(
            srcpkg_name="test",
            srcpkg_version="1.0-1",
            deb_fields={
                "Package": "test-bin",
                "Version": "1.0-1+b1",
                "Architecture": "amd64",
            },
            deb_control_files=[],
        )

        self.assertEqual(
            data_models.get_architectures(artifact_data), {"amd64"}
        )

    def test_DebianPackageBuildLog(self) -> None:
        data = data_models.DebianPackageBuildLog(
            source="test",
            version="1.0-1",
            architecture="amd64",
            filename="test_1.0-1_amd64.buildlog",
        )
        self.assertEqual(data_models.get_architectures(data), {"amd64"})

    def test_DebianBinaryPackage(self) -> None:
        data = data_models.DebianBinaryPackage(
            srcpkg_name="test",
            srcpkg_version="1.0-1",
            deb_fields={
                "Package": "test-bin",
                "Version": "1.0-1+b1",
                "Architecture": "amd64",
            },
            deb_control_files=[],
        )
        self.assertEqual(data_models.get_architectures(data), {"amd64"})

    def test_DebianUpload(self) -> None:
        data = data_models.DebianUpload(
            type="dpkg",
            changes_fields={
                "Source": "hello",
                "Architecture": "amd64",
                "Files": [{"name": "hello_1.0_amd64.deb"}],
            },
        )
        self.assertEqual(data_models.get_architectures(data), {"amd64"})

    def test_mixed_DebianUpload(self) -> None:
        data = data_models.DebianUpload(
            type="dpkg",
            changes_fields={
                "Source": "hello",
                "Architecture": "amd64 all source",
                "Files": [
                    {"name": "hello_1.0.dsc"},
                    {"name": "hello_1.0_amd64.deb"},
                    {"name": "hello-doc_1.0_all.deb"},
                ],
            },
        )
        self.assertEqual(data_models.get_architectures(data), {"amd64", "all"})

    def test_invalid_type(self) -> None:
        data = data_models.EmptyArtifactData()

        with self.assertRaisesRegex(
            TypeError, "^Unexpected type: EmptyArtifactData$"
        ):
            data_models.get_architectures(data)  # type: ignore[arg-type]



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Test the Worker Pool models."""

from django.test import SimpleTestCase

from debusine.server.worker_pools import (
    AWSEC2InstanceRequirements,
    AWSEC2LaunchTemplate,
    AWSEC2NetworkInterface,
    DummyWorkerPoolSpecification,
    MinMaxRequirements,
    worker_pool_specifications_model,
)


class TestWorkerPoolSpecifications(SimpleTestCase):
    """Tests for WorkerPoolSpecifications."""

    def test_worker_pool_specifications_model(self) -> None:
        model = DummyWorkerPoolSpecification(features=["something"])
        result = worker_pool_specifications_model(model.model_dump())
        self.assertEqual(model, result)


class TestAWSEC2LaunchTemplate(SimpleTestCase):
    """Tests for AWSEC2LaunchTemplate."""

    def test_no_instance_type_spec(self) -> None:
        with self.assertRaisesRegex(
            ValueError,
            r"InstanceType or InstanceRequirements must be specified",
        ):
            AWSEC2LaunchTemplate(
                ImageId="my-image",
                InstanceType=None,
                InstanceRequirements=None,
            )

    def get_instance_requirements(self) -> AWSEC2InstanceRequirements:
        """Generate some test instance requirements."""
        return AWSEC2InstanceRequirements(
            VCpuCount=MinMaxRequirements(Min=2),
            MemoryMiB=MinMaxRequirements(Min=1024),
        )

    def test_both_instance_type_specs(self) -> None:
        with self.assertRaisesRegex(
            ValueError,
            r"If InstanceType is specified, InstanceRequirements cannot be",
        ):
            AWSEC2LaunchTemplate(
                ImageId="my-image",
                InstanceType="t3a.medium",
                InstanceRequirements=self.get_instance_requirements(),
            )

    def test_instance_requirements(self) -> None:
        AWSEC2LaunchTemplate(
            ImageId="my-image",
            InstanceRequirements=self.get_instance_requirements(),
        )

    def test_instance_type(self) -> None:
        AWSEC2LaunchTemplate(
            ImageId="my-image",
            InstanceType="t3a.medium",
        )

    def test_network_interfaces_empty_list(self) -> None:
        launch_template = AWSEC2LaunchTemplate(
            ImageId="my-image",
            InstanceType="t3a.medium",
            NetworkInterfaces=[],
        )
        self.assertIsNone(launch_template.NetworkInterfaces)

    def test_network_interfaces_one(self) -> None:
        AWSEC2LaunchTemplate(
            ImageId="my-image",
            InstanceType="t3a.medium",
            NetworkInterfaces=[
                AWSEC2NetworkInterface(
                    AssociatePublicIpAddress=True,
                    DeleteOnTermination=True,
                    DeviceIndex=0,
                    Ipv6AddressCount=1,
                    SubnetId="subnet-42",
                ),
            ],
        )

    def test_network_interfaces_two(self) -> None:
        AWSEC2LaunchTemplate(
            ImageId="my-image",
            InstanceType="t3a.medium",
            NetworkInterfaces=[
                AWSEC2NetworkInterface(
                    AssociatePublicIpAddress=True,
                    DeleteOnTermination=True,
                    DeviceIndex=0,
                    Ipv6AddressCount=1,
                    SubnetId="subnet-42",
                ),
                AWSEC2NetworkInterface(
                    DeviceIndex=1,
                    SubnetId="subnet-43",
                ),
            ],
        )

    def test_network_interfaces_device_indices(self) -> None:
        with self.assertRaisesRegex(
            ValueError, r"Each NetworkInterface's DeviceIndex must be unique"
        ):
            AWSEC2LaunchTemplate(
                ImageId="my-image",
                InstanceType="t3a.medium",
                NetworkInterfaces=[
                    AWSEC2NetworkInterface(
                        DeviceIndex=0,
                    ),
                    AWSEC2NetworkInterface(
                        DeviceIndex=0,
                    ),
                ],
            )

    def test_network_interfaces_groups(self) -> None:
        self.assertIsNone(AWSEC2NetworkInterface(Groups=None).Groups)
        self.assertIsNone(AWSEC2NetworkInterface(Groups=[]).Groups)
        self.assertEqual(
            AWSEC2NetworkInterface(Groups=["sg-01"]).Groups, ["sg-01"]
        )



# Generated by Django 4.2.20 on 2025-05-05 19:20

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('db', '0002_workrequest_aborted_by'),
    ]
    replaces = [
        ('db', '0144_add_workspace_role_viewer'),
    ]

    operations = [
        migrations.AlterField(
            model_name='workspacerole',
            name='role',
            field=models.CharField(
                choices=[
                    ('owner', 'Owner'),
                    ('contributor', 'Contributor'),
                    ('viewer', 'Viewer'),
                ],
                max_length=16,
            ),
        ),
    ]



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Data migration to fix layout of discriminated unions in Key.private_key."""

from django.db import migrations
from django.db.backends.base.schema import BaseDatabaseSchemaEditor
from django.db.migrations.state import StateApps
from django.db.models import Q


def update_key(
    apps: StateApps, schema_editor: BaseDatabaseSchemaEditor
) -> None:
    """Migrate `Key.private_key` to the new layout."""
    Key = apps.get_model("signing", "Key")
    for key in Key.objects.filter(private_key__has_key="data"):
        assert set(key.private_key) == {"storage", "data"}
        data = key.private_key["data"]
        assert "storage" not in data
        key.private_key = {"storage": key.private_key["storage"], **data}
        key.save()


def revert_key(
    apps: StateApps, schema_editor: BaseDatabaseSchemaEditor
) -> None:
    """Migrate `Key.private_key` back to the old layout."""
    Key = apps.get_model("signing", "Key")
    for key in Key.objects.filter(~Q(private_key__has_key="data")):
        data = dict(key.private_key)
        del data["storage"]
        key.private_key = {"storage": key.private_key["storage"], "data": data}
        key.save()


class Migration(migrations.Migration):
    dependencies = [
        ('signing', '0001_initial'),
    ]

    operations = [migrations.RunPython(update_key, revert_key)]



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""DB-aware mock field resolver."""

from debusine.db.models import Artifact, Collection
from debusine.db.tasks.models import DBArtifactInfo, DBInputCollectionSingle
from debusine.tasks.inputs.mock_resolver import (
    MockFieldResolver as BaseMockFieldResolver,
)


class MockFieldResolver(BaseMockFieldResolver):
    """DB-aware mock field resolver."""

    def set_artifact(self, name: str, artifact: Artifact) -> None:
        """Set the value of an input field for an artifact lookup."""
        self.field_values[name] = DBArtifactInfo.from_artifact(
            artifact, label=name, lookup=artifact.id
        )

    def set_collection(self, name: str, collection: Collection) -> None:
        """Set the value of an input field for a collection lookup."""
        self.field_values[name] = DBInputCollectionSingle.from_collection(
            collection, label=name, lookup=collection.id
        )



"""
PostgreSQL settings.

Defaults to unix socket with user auth.
"""

import getpass

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.postgresql',
        'NAME': 'debusine',
        'USER': getpass.getuser(),
        'PASSWORD': '',
        'HOST': '',
        'PORT': '',
        'TEST': {'NAME': 'debusine-test'},
        'ATOMIC_REQUESTS': True,
    }
}



"""Appropriate settings to run the test suite."""

from debusine.project.settings import defaults
from debusine.project.settings.development import *  # noqa: F401, F403

# Don't use bcrypt to run tests (speed gain)
PASSWORD_HASHERS = ['django.contrib.auth.hashers.MD5PasswordHasher']

# Restore INSTALLED_APPS and MIDDLEWARE from defaults to disable debug_toolbar
INSTALLED_APPS = defaults.INSTALLED_APPS.copy()
MIDDLEWARE = defaults.MIDDLEWARE

TEST_NON_SERIALIZED_APPS = ['django.contrib.contenttypes']

# Some checks need to know if we are running tests or not
# Test run with DEBUG=False but it is allowed, for example, to use
# the default SECRET_KEY during tests
TEST_MODE = True

# In normal operation of Debusine, when a Worker / WorkRequest changes,
# it will try to schedule the execution.
# In order to be able to call the scheduler explicitly from the unit tests,
# the automatic scheduling can be disabled. This is only for unit tests
# that need to call the scheduler and verify the results.
DISABLE_AUTOMATIC_SCHEDULING = False

# When running the test suite, enable all apps so that we have all the models
INSTALLED_APPS.extend([])

# Ensure FQDN is an FQDN for use in email addresses
if "." not in defaults.DEBUSINE_FQDN:
    DEBUSINE_FQDN = f"{defaults.DEBUSINE_FQDN}.example.net"

# InMemoryChannelLayer is faster and more reliable in unit tests, but
# unsuitable for production use since it can't communicate between
# processes.
CHANNEL_LAYERS = {
    "default": {"BACKEND": "channels.layers.InMemoryChannelLayer"}
}

# In-memory Redis API implementation, that doesn't require a server.
PROMETHEUS_REDIS_URL = "fakeredis:///1"
